'''
A module to prepare derived arrays for the snapshot.

Examples:
    >>> from ..environment import module_dir
    >>> rules = read_derived_rules([module_dir+'config/derived.cfg'])
    reading config file "pygad/config/derived.cfg"
    >>> assert rules == _rules
    >>> rules['r'], rules['metallicity']
    ('dist(pos)', 'metals/elements.sum(axis=1)')
    >>> general
    {'always_cache': set(['Ekin', 'temp', 'age', 'mag*', 'angmom', 'LX', 'jcirc']), 'cache_derived': True}
    >>> iontable
    {'ions': [], 'tabledir': 'pygad/iontbls/tbls-i45', 'style': 'Oppenheimer new', 'selfshield': False, 'pattern': 'lt<z>f100_i45', 'flux_factor': 1.0, 'T_vals': [2.5, 0.05, 140], 'nH_vals': [-8, 0.05, 160]}

    >>> # from snapshot import Snapshot
    >>> s = Snapshot(module_dir+'snaps/snap_M1196_4x_470')
    >>> ptypes_and_deps(rules['r'], s)
    ([True, True, True, True, True, False], set(['pos']))
    >>> ptypes_and_deps(rules['metallicity'], s) # derived from mass (for all
    ...                                          # particles types) and elements
    ...                                          # (baryons only)
    ([True, False, False, False, True, False], set(['elements', 'metals']))
    >>> s.gas['CIV'] # doctest: +ELLIPSIS
    load block Z... done.
    derive block elements... done.
    derive block H... done.
    load block mass... done.
    load block rho... done.
    load block u... done.
    load block ne... done.
    derive block temp... done.
    derive block CIV... load tables:
      "pygad/iontbls/tbls-i45/lt00000f100_i45" (z=0.000)
      "pygad/iontbls/tbls-i45/lt00491f100_i45" (z=0.049)
    derive block nH... done.
    derive block C... done.
    done.
    SimArr([...],
           units="1e+10 Msol h_0**-1", snap="snap_M1196_4x_470":gas)
    >>> s.gas['MgII'].sum() # doctest: +ELLIPSIS
    derive block MgII... load tables:
      "pygad/iontbls/tbls-i45/lt00000f100_i45" (z=0.000)
      "pygad/iontbls/tbls-i45/lt00491f100_i45" (z=0.049)
    derive block Mg... done.
    done.
    UnitArr(0.00131..., units="1e+10 Msol h_0**-1")
    >>> s.stars['lum'].sum() # doctest: +ELLIPSIS
    load block form_time... done.
    derive block age... done.
    ...
    derive block lum... done.
    UnitArr(3.460675e+10, units="Lsol")

'''
__all__ = ['ptypes_and_deps', 'read_derived_rules', 'general']

from configparser import SafeConfigParser
from .. import utils
from .. import gadget
from .. import environment
from ..units import UnitQty
import re
import warnings
from . import derive_rules
import os

_rules = {}
general = {
        'cache_derived': True,
        'always_cache': set(),
}
iontable = {
        'tabledir':     None,
        'pattern':      'lt<z>f100_i45',
        'style':        'Oppenheimer new',
        'flux_factor':  1.0,
        'selfshield':   False,
        'ions':         [],
        'nH_vals':      [-8  , 0.05, 160],
        'T_vals':       [ 2.5, 0.05, 140],
}

def ptypes_and_deps(defi, snap):
    '''
    Get the ptypes and dependecies of a block definition.

    The ptypes are the greates common set of particle types of all the
    dependencies.

    Args:
        defi (str):     The definition of the block in form of a expression (that
                        is interpretable by Snap.get).
        snap (Snap):    The snapshot the block shall be added to.

    Returns:
        ptypes (list):  A list of length 6 with booleans for the particle types.
        deps (set):     The set of the names of the blocks it depends on.

    Raises:
        ValueError
    '''
    ptypes = [True] * 6    # gets restricted in the following
    deps = set()
    root = snap.root
    for name in utils.iter_idents_in_expr(defi):
        if name in root._block_avail:
            ptypes = [(pt and avail) for pt,avail
                        in zip(ptypes,root._block_avail[name])]
            deps.add(name)
        elif name in gadget.families:
            fam = gadget.families[name]
            ptypes = [(ptypes[i] and (i in fam)) for i in range(6)]
        elif hasattr(derive_rules, name):
            func = getattr(derive_rules, name)
            if not hasattr(func,'_deps') and not func is UnitQty:
                warnings.warn('The derived block defining function ' +
                              '"%s" has not attribute `_deps` ' % name +
                              'defining its dependencies! -- Assume no ' +
                              'dependencies.')
            for dep in getattr(func,'_deps',set()):
                if dep in root._block_avail:
                    ptypes = [(pt and avail) for pt,avail
                                in zip(ptypes,root._block_avail[dep])]
                    deps.add(dep)
                else:
                    # all dependencies are needed -> we can shortcut
                    return [False]*6, set()
    return ptypes, deps

def read_derived_rules(config, delete_old=False):
    '''
    Read rules for derived blocks from a config file.

    The config file consists of a single section 'rules' that has names and their
    corresponding rules (Python expressions, that have the namespace of the
    snapshot and numpy -- see Snap.get for more information).

    Args:
        config (list):              list of possible filenames for the config file.
        delete_old (bool):          Delete old definition before add the new ones.
    '''
    global _rules, general, iontable

    def test_section(cfg, section, entries):
        if not cfg.has_section(section):
            raise KeyError('Section "%s" is required in config ' % section +
                           'file for derived blocks.')
        if set(cfg.options(section)) < set(entries):
            raise ValueError('Section "%s" must have the ' % section +
                             'following entries: ' + str(entries))

    from os.path import exists, expanduser
    for filename in config:
        if exists(expanduser(filename)):
            break
    else:
        raise IOError('Config file "%s" does not exist!' % config)

    if environment.verbose >= environment.VERBOSE_NORMAL:
        print('reading config file "%s"' % filename)

    # The SafeConfigParser class has been renamed to ConfigParser in Python 3.2, not yet removed for compatibility
    cfg = SafeConfigParser(allow_no_value=True,
                           inline_comment_prefixes=('#', ';'))  # new to python3. ignores comments at end of values
    cfg.optionxform = str
    cfg.read(filename)

    test_section(cfg, 'general', ['cache_derived'])
    test_section(cfg, 'rules', [])

    general['cache_derived'] = cfg.getboolean('general', 'cache_derived')
    if cfg.has_option('general', 'always_cache'):
        general['always_cache'] = set(block.strip() for block
                in cfg.get('general','always_cache').split(','))

    if delete_old:
        _rules.clear()
        general.clear()
        general['cache_derived'] = True
        general['always_cache'] = set()
        iontable.clear()
        iontable['tabledir'] = None
        iontable['pattern'] = 'lt<z>f100_i31'
        iontable['style'] = 'Oppenheimer new'
        iontable['flux_factor'] = 1.0
        iontable['selfshield'] = False
        iontable['ions'] = []
        iontable['nH_vals'] = [-8  , 0.05, 160]
        iontable['T_vals']  = [ 2.5, 0.05, 140]

    if cfg.has_section('iontable'):
        #test_section(cfg, 'iontable', ['tabledir', 'ions', 'nH_vals', 'T_vals'])
        test_section(cfg, 'iontable', ['tabledir', 'pattern', 'style'])
        iontable['tabledir'] = cfg.get('iontable', 'tabledir',
                                       vars={'PYGAD_DIR':environment.module_dir})
        iontable['pattern'] = cfg.get('iontable', 'pattern',
                                      vars={'PYGAD_DIR':environment.module_dir})
        iontable['style'] = cfg.get('iontable', 'style')
        if cfg.has_option('iontable', 'ions'):
            iontable['ions'] = [ion.strip()
                    for ion in cfg.get('iontable','ions').split(',')]
        for vals in ['nH_vals', 'T_vals']:
            if cfg.has_option('iontable', vals):
                iontable[vals] = [float(v.strip())
                        for v in cfg.get('iontable',vals).split(',')]
        if cfg.has_option('iontable', 'selfshield'):
            iontable['selfshield'] = cfg.getboolean('iontable', 'selfshield')
        if cfg.has_option('iontable', 'flux_factor'):
            iontable['flux_factor'] = cfg.getfloat('iontable', 'flux_factor')

    for i,el in enumerate(gadget.elements):
        _rules[el] = 'elements[:,%d]' % i
    _rules.update( cfg.items('rules') )

    for derived_name in list(_rules.keys()):
        if derived_name=='lum':
            mag, lum = 'mag', 'lum'
        elif re.match('lum_[a-zA-Z]', derived_name):
            mag, lum = 'mag_'+derived_name[-1], derived_name
        else:
            continue
        _rules[mag] = "lum_to_mag(%s)" % lum

    # add the ions from the Cloudy table as derived blocks
    if os.path.isdir(iontable['tabledir']):
        if iontable['style'] == 'Oppenheimer new':
            from ..cloudy.cloudy_tables import config_ion_table
            verbose = environment.verbose
            environment.verbose = environment.VERBOSE_QUIET
            ions = config_ion_table(0.0).ions
            environment.verbose = verbose
        else:
            ions = iontable['ions']
        for ion in ions:
            el, ionisation = ion.split()
            ion = el + ionisation   # getting rid of the white space
            if ion in _rules:
                continue
            _rules[ion] = "calc_ion_mass(gas, '%s', '%s', selfshield=%s)" % (
                                    el, ionisation, iontable['selfshield'])
    else:
        print("Warning: no ion table found in: " + iontable['tabledir'] + ";")
        print("         ion blocks (e.g. HI) will not be available")

    return _rules




'''
Terminal helper functions.
'''
__all__ = ['strip_ansi', 'term_len', 'get_terminal_size']

import re
import os

_ansi_re = re.compile('\033\[((?:\d|;)*)([a-zA-Z])')


def strip_ansi(value):
    return _ansi_re.sub('', value)


def term_len(txt):
    '''Determine the length of the text, if printed to the terminal.'''
    return len(strip_ansi(txt))


def get_terminal_size():
    '''
    Determine the size of the terminal.

    Uses `os.popen('stty size', 'r')`. Should work on any Unix system.

    Returns:
        columns (int):  The number of columns.
        rows (int):     The number of rows.
    '''
    rows, columns = os.popen('stty size', 'r').read().split()
    return int(columns), int(rows)




"""
Module for snapshot analysis.

Also doctest other parts of this sub-module:
    >>> import doctest
    >>> doctest.testmod(sph_eval)
    TestResults(failed=0, attempted=15)
    >>> doctest.testmod(properties)
    TestResults(failed=0, attempted=34)
    >>> doctest.testmod(halo)
    TestResults(failed=0, attempted=29)
    >>> doctest.testmod(profiles)
    TestResults(failed=0, attempted=19)
    >>> doctest.testmod(absorption_spectra)
    TestResults(failed=0, attempted=17)

    #>>> doctest.testmod(analysis)
    #TestResults(failed=0, attempted=20)
"""
from .sph_eval import *
from .properties import *
from .halo import *
from .profiles import *

# from analysis import *
from .absorption_spectra import *
from .vpfit import *



'''
Some general quantities and constants.

Examples:
    >>> solar.Z()
    0.012946810000000086
    >>> assert abs(sum(solar.Z_massfrac) - 1.0) < 1e-3
    >>> c
    UnitArr(2.997925e+08, units="m s**-1")
    >>> if abs(G - '6.67408e-11 m**3/kg/s**2') > '1e-14 m**3/kg/s**2':
    ...     print(G)
    >>> if abs(kB - '1.380649e-23 J/K') > '2e-29 J/K':
    ...     print(kB)
    >>> if abs(N_A - '6.022141e23 mol**-1') > '2e17 mol**-1':
    ...     print(N_A)
    >>> if abs(m_u - '1.660539e-27 kg') > '2e-33 kg':
    ...     print(m_u)
    >>> for m in [m_p, m_n, 1822.9*m_e]:
    ...     assert abs(m/m_u - 1.0) < 0.02

    TODO: Why is this actually not closer to one?!
'''
__all__ = ['alpha_elements',
           'constants',
           'G', 'c', 'h', 'hbar', 'kB', 'N_A', 'epsilon0', 'mu0', 'q_e', 'm_p',
           'm_n', 'm_u', 'm_e', 'm_H', 'm_He', 'R',
           'solar',
           'SMH_Moster_2013', 'SMH_Behroozi_2013', 'SMH_Kravtsov_2014',
           'Reff_van_der_Wel_2014', 'SFR_Elbaz_2007',
           'Jeans_length', 'Jeans_mass']

import numpy as np
import warnings
import scipy.constants
from ..units import *
import sys

# alpha elements, produced in (or actually before) SNII:
alpha_elements = ['O', 'C', 'Ne', 'Si', 'Mg', 'S', 'Ca']
# sometimes Carbon & Nitrogen are also considered alpha elements; sometimes Oxygen
# is not considered an alpha element; Ar and Ti are not followed...

# All scipy constants:
constants = {}
for name in scipy.constants.find():
    try:
        value = scipy.constants.value(name)
        units = scipy.constants.unit(name)
        constants[name] = UnitArr(value,
                                  units.replace('^', '**').replace('ohm', 'Ohm')
                                  if units else None)
    except UnitError:
        pass
    except SyntaxError:
        pass
# Some useful constants:
try:
    G = constants['Newtonian constant of gravitation']
    c = constants['speed of light in vacuum']
    h = constants['Planck constant']
    hbar = constants['reduced Planck constant']
    kB = constants['Boltzmann constant']
    N_A = constants['Avogadro constant']
    epsilon0 = constants['vacuum electric permittivity']
    mu0 = constants['vacuum mag. permeability']
    q_e = constants['elementary charge']
    m_p = constants['proton mass']
    m_n = constants['neutron mass']
    m_u = constants['atomic mass constant']
    m_e = constants['electron mass']
    m_H = UnitArr(1.00794, 'u')
    m_He = UnitArr(4.002602, 'u')
    R = N_A * kB  # ideal gas constant
except KeyError:
    print("Constant missing from scipy.constants, update scipy to the latest"\
          " version!")
    raise

class solar(object):
    '''
    Class serving as a structure for holding solar values.

    These are taken from the tables from Wiersma+(2008) and are the CLOUDY
    default values.
    webpage: http://www.strw.leidenuniv.nl/WSS08/
    The link above is dead, alternative source on our bitbucket: https://bitbucket.org/broett/pygad/downloads/
    '''

    Z_massfrac = [0.70649785,
                  0.28055534,
                  0.0020665436,
                  8.3562563e-4,
                  0.0054926244,
                  0.0014144605,
                  5.907064E-4,
                  6.825874E-4,
                  4.0898522E-4,
                  6.4355E-5,
                  0.0011032152]

    abs_mag = 4.75

    @staticmethod
    def H():
        '''The Hydrogen mass fraction.'''
        return solar.Z_massfrac[0]

    @staticmethod
    def He():
        '''The Helium mass fraction.'''
        return solar.Z_massfrac[1]

    @staticmethod
    def Z():
        '''The solar metallicty.'''
        return 1.0 - (solar.Z_massfrac[0] + solar.Z_massfrac[1])

    @staticmethod
    def C():
        '''The Carbon mass fraction.'''
        return solar.Z_massfrac[2]

    @staticmethod
    def N():
        '''The Nitrogen mass fraction.'''
        return solar.Z_massfrac[3]

    @staticmethod
    def O():
        '''The Oxygen mass fraction.'''
        return solar.Z_massfrac[4]

    @staticmethod
    def Ne():
        '''The Neon mass fraction.'''
        return solar.Z_massfrac[5]

    @staticmethod
    def Mg():
        '''The Magnesium mass fraction.'''
        return solar.Z_massfrac[6]

    @staticmethod
    def Si():
        '''The Silicon mass fraction.'''
        return solar.Z_massfrac[7]

    @staticmethod
    def S():
        '''The Sulfur mass fraction.'''
        return solar.Z_massfrac[8]

    @staticmethod
    def Ca():
        '''The Calcium mass fraction.'''
        return solar.Z_massfrac[9]

    @staticmethod
    def Fe():
        '''The Iron mass fraction.'''
        return solar.Z_massfrac[10]

    @staticmethod
    def alpha():
        '''The mass fraction of the alpha elements.'''
        return sum(getattr(solar, e)() for e in alpha_elements)

    @staticmethod
    def log10_Fe_H():
        '''log_10(Fe / H)'''
        return np.log10(solar.Fe() / solar.H())

    @staticmethod
    def log10_O_Fe():
        '''log_10(O/ Fe)'''
        return np.log10(solar.O() / solar.Fe())

    @staticmethod
    def log10_alpha_Fe():
        '''log_10(alpha-elements / Fe)'''
        return np.log10(solar.alpha() / solar.Fe())


def SMH_Moster_2013(M_halo, z=0.0, return_scatter=False):
    '''
    Calculate the stellar mass to halo mass ratio as in Moster et al. (2013).

    The formula was given by Benjamin Moster in a private communication. It simply
    comes from a simple propagation of uncertainties of the formula of the mean,
    plus a constant 0.15 dex for a "physical scatter" (as explained in the paper).

    Args:
        M_halo (UnitQty):       The halo mass (in solar masses, if as float).
        z (float):              The redshift to calculate for.
        return_scatter (bool):  Whether to also return the scatter.

    Returns:
        SMH (float, np.dnarray):    The stellar mass to halo mass ratio.
       [lower (float, np.dnarray):  The lower bound of the 1-sigma scatter.]
       [upper (float, np.dnarray):  The upper bound of the 1-sigma scatter.]

    Example:
        >>> SMH_Moster_2013([1e11, 1e12, 1e13])
        array([0.01012983, 0.03429659, 0.00974028])
        >>> for a in SMH_Moster_2013([1e11, 1e12, 1e13], return_scatter=True):
        ...     print(a)
        [0.01012983 0.03429659 0.00974028]
        [0.00341559 0.0186693  0.00423019]
        [0.03004267 0.06300485 0.0224276 ]
        >>> for z in [0.5, 1.0, 2.0, 4.0]:
        ...     print(SMH_Moster_2013('1e11 Msol',z,return_scatter=True))
        ...     print(SMH_Moster_2013('1e13 Msol',z,return_scatter=True))
        (0.004314181996911216, 0.0013140763985384505, 0.0141636866191142)
        (0.009955587054568093, 0.004043156124677078, 0.02451394666561377)
        (0.003235536108623059, 0.0008746580727273953, 0.011968898746409244)
        (0.010323427157890868, 0.003921558687759114, 0.027176221693924904)
        (0.0026531006668289547, 0.0006021546514267902, 0.011689593581399173)
        (0.010560160483693697, 0.0037325186267171905, 0.02987714211072727)
        (0.002390900353994381, 0.00044896686807466534, 0.012732352672802963)
        (0.0103622356309934, 0.0034261919431899795, 0.03133972907900081)
    '''
    param = {
        'M1': [11.590470, 1.194913],
        'N': [0.035113, -0.024729],
        'beta': [1.376177, -0.825820],
        'gamma': [0.608170, 0.329275],
    }
    one_minus_a = float(z / (z + 1.0))

    M_halo = UnitQty(M_halo, 'Msol', dtype=float)
    from collections import Iterable
    if getattr(M_halo, 'shape', None) and M_halo.shape[0] > 1:
        ret = []
        for Mh in M_halo:
            ret.append(SMH_Moster_2013(Mh, z=z, return_scatter=return_scatter))
        ret = np.array(ret)
        if return_scatter:
            return ret[:, 0], ret[:, 1], ret[:, 2]
        else:
            return ret
    M_halo = float(M_halo)

    inter = {}
    for key, vals in param.items():
        inter[key] = vals[0] + vals[1] * one_minus_a
    inter['M1'] = 10 ** inter['M1']

    SMH = 2.0 * inter['N'] / ((M_halo / inter['M1']) ** -inter['beta'] +
                              (M_halo / inter['M1']) ** inter['gamma'])
    if return_scatter:
        param.update({
            'M1e': [0.236067, 0.353477],
            'Ne': [0.00577173, 0.00693815],
            'betae': [0.19344, 0.285018],
            'gammae': [0.0993274, 0.212919],
        })

        eta = M_halo / inter['M1']
        alpha = eta ** -inter['beta'] + eta ** inter['gamma']
        dmd = {}
        dmd['M1'] = [(inter['gamma'] * eta ** inter['gamma'] \
                      - inter['beta'] * eta ** (-inter['beta'])) / alpha,
                     (inter['gamma'] * eta ** inter['gamma'] \
                      - inter['beta'] * eta ** (-inter['beta'])) / alpha \
                     * one_minus_a]
        dmd['N'] = [np.log10(np.e) / inter['N'],
                    np.log10(np.e) / inter['N'] * one_minus_a]
        dmd['beta'] = [np.log10(np.e) / alpha * np.log(eta) \
                       * eta ** (-inter['beta']),
                       np.log10(np.e) / alpha * np.log(eta) \
                       * eta ** (-inter['beta']) * one_minus_a]
        dmd['gamma'] = [-np.log10(np.e) / alpha * np.log(eta) \
                        * eta ** inter['gamma'],
                        -np.log10(np.e) / alpha * np.log(eta) \
                        * eta ** inter['gamma'] * one_minus_a]
        sigma = 0.0
        for key in dmd.keys():
            sigma += (dmd[key][0] * param[key + 'e'][0]) ** 2 \
                     + (dmd[key][1] * param[key + 'e'][1]) ** 2
        sigma = np.sqrt(sigma) + 0.15

        log10_SM = np.log10(SMH * M_halo)
        lower, upper = 10 ** (log10_SM - sigma) / M_halo, 10 ** (log10_SM + sigma) / M_halo

    if return_scatter:
        return SMH, lower, upper
    else:
        return SMH


def _Behroozi_function(log10_M, log10_M1, log10_eps, alpha, delta, gamma):
    '''
    The function form from Behroozi et al. (2013) for calculating the logarithm
    of stellar mass as a function of the logarithm of halo mass.
    '''

    def f(x, alpha, delta, gamma):
        return -np.log10(10 ** (alpha * x) + 1.) + \
               delta * np.log10(1 + np.exp(x)) ** gamma / (1 + np.exp(10 ** -x))

    return log10_eps + log10_M1 + f(log10_M - log10_M1, alpha, delta, gamma) \
           - f(0, alpha, delta, gamma)


def SMH_Behroozi_2013(M_halo, z=0.0, return_scatter=False):
    '''
    Calculate the stellar mass to halo mass ratio as in Behroozi et al. (2013).

    Args:
        M_halo (UnitQty):       The halo mass (in solar masses, if as float).
        z (float):              The redshift to calculate for.
        return_scatter (bool):  Whether to also return the scatter.

    Returns:
        SMH (float, np.dnarray):    The stellar mass to halo mass ratio.
       [lower (float, np.dnarray):  The lower bound of the 1-sigma scatter.]
       [upper (float, np.dnarray):  The upper bound of the 1-sigma scatter.]

    Example:
        >>> SMH_Behroozi_2013([1e11, 1e12, 1e13])
        array([0.0046755 , 0.02679825, 0.00898848])
        >>> for e in SMH_Behroozi_2013([1e11, 1e12, 1e13], return_scatter=True):
        ...     print(e)
        [0.0046755  0.02679825 0.00898848]
        [0.00283027 0.01622207 0.00544109]
        [0.00772375 0.04426968 0.01484862]
        >>> for z in [0.5, 1.0, 2.0, 4.0]:
        ...     print(SMH_Behroozi_2013(1e11,z))
        ...     print(SMH_Behroozi_2013(1e13,z))
        0.004210670265291123
        0.009949059926042463
        0.0031404619764022197
        0.010782128125330971
        0.002244488003457852
        0.011437102413458175
        0.003460606635214448
        0.007272677770327794
    '''
    M_halo = UnitQty(M_halo, 'Msol', dtype=float).view(np.ndarray)

    z = float(z)
    a = 1 / (1. + z)
    nu = np.exp(-4. * a ** 2)
    log10_eps = -1.777 + (-0.006 * (a - 1.) - 0.000 * z) * nu - 0.119 * (a - 1.)
    log10_M1 = 11.514 + (-1.793 * (a - 1.) - 0.251 * z) * nu
    alpha = -1.412 + (0.731 * (a - 1.)) * nu
    delta = 3.508 + (2.608 * (a - 1.) - 0.043 * z) * nu
    gamma = 0.316 + (1.319 * (a - 1.) + 0.279 * z) * nu

    mu = -0.020 + 0.081 * (a - 1.)
    kappa = 0.045 - 0.155 * (a - 1.)

    xi = 0.218 - 0.023 * (a - 1.)

    log10_SM = _Behroozi_function(np.log10(M_halo), log10_M1, log10_eps, alpha, delta, gamma)
    # TODO: do properly
    # log10_SM = log10_SM - mu
    if return_scatter:
        return 10 ** log10_SM / M_halo, \
               10 ** (log10_SM - xi) / M_halo, 10 ** (log10_SM + xi) / M_halo
    else:
        return 10 ** log10_SM / M_halo


def SMH_Kravtsov_2014(M_halo, type='200c', return_scatter=False):
    '''
    Calculate the stellar mass to halo mass ratio as in Kravtsov et al. (2014).

    Note:
        The fits themselves are in Appendix A.

    Args:
        M_halo (UnitScalar):    The halo mass (in solar masses, if as float).
        return_scatter (bool):  Whether to also return the scatter.

    Returns:
        SMH (float):                The stellar mass to halo mass ratio.
       [scatter ([float, float]):   The scatter.]

    Example:
        >>> SMH_Kravtsov_2014([1e11, 1e12, 1e13])
        array([0.00807214, 0.04218667, 0.0205807 ])
        >>> for a in SMH_Kravtsov_2014([1e11, 1e12, 1e13], return_scatter=True):
        ...     print(a)
        [0.00807214 0.04218667 0.0205807 ]
        [0.00509317 0.02661799 0.01298554]
        [0.01279348 0.06686137 0.03261821]
        >>> for M in [1e11, 1e12, 1e13, 1e14, 1e15]:
        ...     print(SMH_Kravtsov_2014(M, return_scatter=True))
        (0.008072138072586098, 0.005093174802556399, 0.01279347667984786)
        (0.042186669544032114, 0.02661798898796599, 0.06686136537294297)
        (0.02058069982657301, 0.012985543710118501, 0.03261821105122113)
        (0.007212575378877716, 0.004550827407919915, 0.01143116161809591)
        (0.0022639956645691835, 0.0014284846924312448, 0.003588191316537168)
    '''
    M_halo = UnitQty(M_halo, 'Msol', dtype=float).view(np.ndarray)

    log10_M1 = {'200c': 11.35, '200m': 11.41, 'vir': 11.39}[type]
    log10_eps = {'200c': -1.642, '200m': -1.720, 'vir': -1.685}[type]
    alpha = {'200c': -1.779, '200m': -1.727, 'vir': -1.740}[type]
    delta = {'200c': 4.345, '200m': 4.305, 'vir': 4.335}[type]
    gamma = {'200c': 0.619, '200m': 0.544, 'vir': 0.531}[type]

    log10_SM = _Behroozi_function(np.log10(M_halo), log10_M1, log10_eps, alpha, delta, gamma)
    if return_scatter:
        warnings.warn('Scatter of 0.2 dex in Kravtsov et al. (2014) not sure!')
        return 10 ** log10_SM / M_halo, \
               10 ** (log10_SM - 0.2) / M_halo, 10 ** (log10_SM + 0.2) / M_halo
    else:
        return 10 ** log10_SM / M_halo


def Reff_van_der_Wel_2014(M_stars, z, type, return_scatter=False):
    '''
    A fit to the measure effective radii from van der Wel et al. (2014).

    Args:
        M_stars (UnitQty):      The stellar mass (in solar masses if float).
        z (float):              The redshift. Has to be less than 3.
        type ('ETG', 'LTG'):    The type of the galaxy.
        return_scatter (bool):  Whether to also return the scatter.

    Returns:
        Reff (UnitQty):     The half mass radius in kpc.
       [sigma (float):      The logarithmic scatter in Reff:
                            sigma(log10(Reff/kpc)). It is also for arrays of Reff
                            just a scalar!]

    Examples:
        >>> Reff_van_der_Wel_2014(1e10, z=1.2, type='LTG')
        UnitArr(3.5828332753571375, units="kpc")
        >>> Reff_van_der_Wel_2014([1e9,1e10,1e11], z=0.5, type='LTG')
        UnitArr([2.63476599, 4.52628659, 7.7757457 ], units="kpc")
        >>> Reff_van_der_Wel_2014([1e10,1e11,1e12], z=0.5, type='ETG',
        ...                       return_scatter=True)
        (UnitArr([ 0.99942888,  5.36725089, 28.82384388], units="kpc"), 0.10500000000000001)
    '''
    z_edges = np.array([0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0])
    z = float(z)
    if z > z_edges[-1]:
        raise ValueError('The redshift has to be less than %.3f.' % z_edges[-1])

    A_bins = {'ETG': [0.60, 0.42, 0.22, 0.09, -0.05, -0.06],
              'LTG': [0.86, 0.78, 0.70, 0.65, 0.55, 0.51]}
    alpha_bins = {'ETG': [0.75, 0.71, 0.76, 0.76, 0.76, 0.79],
                  'LTG': [0.25, 0.22, 0.22, 0.23, 0.22, 0.18]}
    sigma_bins = {'ETG': [0.10, 0.11, 0.12, 0.14, 0.14, 0.14],
                  'LTG': [0.16, 0.16, 0.17, 0.18, 0.19, 0.19]}

    z_bins = (z_edges[1:] + z_edges[:-1]) / 2.0
    for i in range(1, len(z_bins)):
        if z < z_bins[i] or i == len(z_bins) - 1:
            xi_z = (z - z_bins[i]) / (z_bins[i - 1] - z_bins[i])
            A = A_bins[type][i - 1] * xi_z + A_bins[type][i] * (1. - xi_z)
            alpha = alpha_bins[type][i - 1] * xi_z + alpha_bins[type][i] * (1. - xi_z)
            sigma = sigma_bins[type][i - 1] * xi_z + sigma_bins[type][i] * (1. - xi_z)
            break

    M_stars = UnitQty(M_stars, 'Msol', dtype=float) / UnitArr(5e10, 'Msol')
    M_stars = M_stars.view(np.ndarray)  # needed for broken power of alpha
    Reff = UnitArr(10 ** A * M_stars ** alpha, 'kpc')
    if return_scatter:
        return Reff, sigma
    else:
        return Reff


def SFR_Elbaz_2007(M_star, z=0.0, return_scatter=False):
    '''
    The SFR as a function of stellar mass from Elbaz et al. (2007).

    The paper provides fits to observational data at z=0 and z=1.
    Here, it is interpolated between those two values, if z is
    inbetween or >1.

    Args:
        M_star (UnitQty):       The stellar mass of the galaxy.
        z (float):              The redshift. Has to be less than 1.5.
        return_scatter (bool):  Whether to also return the scatter.

    Returns:
        SFR (UnitQty):          The star-formation rate.
       [lower (UnitQty):        The lower bound of the 1-sigma scatter.]
       [upper (UnitQty):        The upper bound of the 1-sigma scatter.]

    Examples:
        >>> SFR_Elbaz_2007('1e9 Msol')
        UnitArr(0.2509107407720147, units="Msol yr**-1")
        >>> SFR_Elbaz_2007([1e10, 1e11])
        UnitArr([1.47747198, 8.7       ], units="Msol yr**-1")
        >>> SFR_Elbaz_2007([1,2e10,5e11], z=1)
        UnitArr([7.20000000e-09, 1.34356751e+01, 2.43447602e+02], units="Msol yr**-1")
        >>> for e in SFR_Elbaz_2007(2e10, z=0.1, return_scatter=True):
        ...     print(e)
        3.611106995357125 [Msol yr**-1]
        1.974967367437606 [Msol yr**-1]
        6.883386251196164 [Msol yr**-1]
        >>> for e in SFR_Elbaz_2007([1,2e10,5e11], z=0.7, return_scatter=True):
        ...     print(e)
        [1.38838325e-08 1.01608191e+01 1.79425865e+02] [Msol yr**-1]
        [7.60266234e-09 5.13688082e+00 9.03862837e+01] [Msol yr**-1]
        [2.64461727e-08 2.02086955e+01 3.57505029e+02] [Msol yr**-1]
    '''
    M_star = UnitQty(M_star, 'Msol', dtype=float).view(np.ndarray)

    SFR_0 = 8.7 * (M_star / 1e11) ** 0.77
    SFR_1 = 7.2 * (M_star / 1e10) ** 0.9

    sigma_0 = [5.0 * (M_star / 1e11) ** 0.77, 16.1 * (M_star / 1e11) ** 0.77]
    sigma_1 = [3.6 * (M_star / 1e10) ** 0.9, 14.4 * (M_star / 1e10) ** 0.9]

    xi_z = (float(z) - 1.0) / (0.0 - 1.0)
    SFR = xi_z * SFR_0 + (1. - xi_z) * SFR_1
    sigma = [xi_z * s[0] + (1. - xi_z) * s[1] for s in zip(sigma_0, sigma_1)]

    SFR = UnitArr(SFR, 'Msol/yr')
    sigma = [UnitArr(s, 'Msol/yr') for s in sigma]

    if return_scatter:
        return SFR, sigma[0], sigma[1]
    else:
        return SFR


def Jeans_length(T, rho, mu='1 u', units='kpc'):
    '''
    The Jeans length.

    ... as in:

        /    15 kB T    \ 1/2
       | --------------- |
        \ 4 pi G mu rho /

    Note:
        You can also calculate Jeans lengthes for arrays of parameters as long as
        they have all the same length (the Jeans length is then calculated
        elementwise for all entries in the array) or only on parameter is given
        as an array while all others are still scalars.

    Args:
        T (UnitArr):        The temperature of the gas (floats are interpreted in
                            units of Kelvin).
        rho (UnitArr):      The gas density (floats -> g/cm**3).
        mu (UnitArr):       The (effective) particle mass (floats -> g).
        units (str, Unit):  The units to return the Jeans length in.

    Return:
        L (UnitQty):        The Jeans length.

    Raises:
        ValueError:         If multiple parameters have nonidentical shapes.

    Examples:
        >>> if abs(Jeans_length('10 K', '1e6 u/cm**3', units='pc') -
        ...        '0.03067 pc') > '1e-4 pc':
        ...     print(Jeans_length('10 K', '1e6 u/cm**3', units='pc'))
        >>> if abs(Jeans_length(1e4, '1 u/cm**3', units='pc') -
        ...        '969.8 pc') > '0.5 pc':
        ...     print(Jeans_length(1e4, '1 u/cm**3', units='pc'))
    '''
    k = []  # array for lengths of given arrays
    T = UnitQty(T, units='K', dtype=np.float64)
    if sum(T.shape) > 0:
        k.append(sum(T.shape))

    rho = UnitQty(rho, units='g/cm**3', dtype=np.float64)
    if sum(rho.shape) > 0:
        k.append(sum(rho.shape))

    mu = UnitQty(mu, units='g', dtype=np.float64)
    if sum(mu.shape) > 0:
        k.append(sum(mu.shape))

    # check if all given arrays are of the same shape, if there are more than one
    if len(k) > 1:
        if k.count(k[0]) != len(k):
            raise ValueError('If more than one parameter is passed as an ' + \
                             'array, they need to have the same shape!')

    L2 = 15. * kB * T / (4. * np.pi * G * mu * rho)
    L2.convert_to(Unit(units) ** 2)
    return np.sqrt(L2)


def Jeans_mass(T, rho, mu='1 u', units='Msol'):
    '''
    The Jeans mass.

    ... as in:

        4      3          /     3    \ 1/2  / 5 kB T \ 3/2
       --- pi L  rho  =  | ---------- |    | -------- |
        3                 \ 4 pi rho /      \  G mu  /

    Note:
        You can also calculate Jeans lengthes for arrays of parameters as long as
        they have all the same length (the Jeans length is then calculated
        elementwise for all entries in the array) or only on parameter is given
        as an array while all others are still scalars.

    Args:
        T (UnitArr):        The temperature of the gas (floats are interpreted in
                            units of Kelvin).
        rho (UnitArr):      The gas density (floats -> g/cm**3).
        mu (UnitArr):       The (effective) particle mass (floats -> g).
        units (str, Unit):  The units to return the Jeans mass in.

    Return:
        M (UnitQty):        The Jeans mass.

    Raises:
        ValueError:         If multiple parameters have nonidentical shapes.

    Examples:
        >>> M = Jeans_mass('10 K', '1e6 u/cm**3')
        >>> if abs(M - '2.964 Msol') > '1e-3 Msol':
        ...     print(M)
        >>> M = Jeans_mass(UnitArr([1e2,1e3,1e4,1e5],'K'), '1.0 u/cm**3')
        >>> if np.any(np.abs(M -
        ...         UnitArr([9.372e+04,2.964e+06,9.372e+07,2.964e+09],'Msol'))
        ...         / M > 1e-3):
        ...     print(M)
        ...     print(np.abs(M -
        ...         UnitArr([9.372e+04,2.964e+06,9.372e+07,2.964e+09],'Msol')) / M)
        >>> M = Jeans_mass(UnitArr([1e4,1e4,1e2],'K'), UnitArr([1e-2,1e0,1e2],'u/cm**3'))
        >>> if np.any(np.abs(M -
        ...         UnitArr([9.372e+08, 9.372e+07, 9.372e+03],'Msol')) / M > 1e-3):
        ...     print(M)
    '''
    rho = UnitQty(rho, units='g/cm**3', dtype=np.float64)

    M = 4. * np.pi / 3. * Jeans_length(T, rho, mu) ** 3 * rho
    M.convert_to(units)
    return M



'''
Functions for kernels that can be derived from the definitions.

Note:
    Outside the kernels' support, the function return undefined values! If
    arbitrary values of radius are use, it has to be taken care of those outside
    the support manually.
                                                                              r
    Furthermore the arguments are always normed by the smoothing length: u = --- .
                                                                             2 h

Examples and doctests:
    Test some properties of the projected kernels:
    >>> f = project_kernel('Wendland C4')
    >>> integ = 0.0; dx=0.1
    >>> for x in np.arange(-1,1,dx):
    ...     for y in np.arange(-1,1,dx):
    ...         r = np.sqrt(x**2+y**2)
    ...         integ += f(r)
    >>> integ *= dx**2
    >>> if not abs(integ-1.0) < 1e-3:
    ...     print(integ)

    Test some properties of the integrated kernels:
    >>> for ikernel in kernels.items():
    ...     kernel = ikernel[0]
    ...     integ_kernel = integrate_kernel(kernel)
    ...     if abs(integ_kernel(1.0) - 1.0) > 1e-6:
    ...         print('Kernel %s did not got integrated to one!', name)
    ...     if abs(integ_kernel(0.0)) > 1e-6:
    ...         print('Integrated kernel %s does not start at zero!', name)
    ...     for r in [0.1, 0.3, 0.5, 0.8, 0.9]:
    ...         assert 0 < integ_kernel(r) < 1

    Test some properties of the random radii:
    >>> N = 1000
    >>> rs = np.empty(N)
    >>> for ikernel in kernels.items():
    ...     kernel = ikernel[0]
    ...     for n in range(N):
    ...         rs[n] = rand_kernel_r(kernel)
    ...     pdf, edges = np.histogram(rs, bins=np.linspace(0,1,30+1))
    ...     integ_rand = np.cumsum(pdf) / float(N)
    ...     mids = (edges[:-1]+edges[1:]) / 2.
    ...     integ_kernel = integrate_kernel(kernel)
    ...     for i,r in enumerate(mids):
    ...         ref_val = integ_kernel(r)
    ...         err = abs(ref_val - integ_rand[i])
    ...         if err>0.1 and err/ref_val>0.1:
    ...             print('Kernel "%s" was wrongly sampled at r=%f: err=%f' % (kernel, r, err))
'''
__all__ = ['project_kernel', 'integrate_kernel', 'rand_kernel_r']

import numpy as np
from .definitions import *
from ..utils import static_vars

def project_kernel(kernel, N=100, inter_kind='quadratic'):
    '''
    Create an interpolation function for the projected kernel.

    Args:
        kernel (str):       The name of the kernel to project (as in
                            kernels.kernel).
        N (int):            The number of points to use for the interpolation.
        inter_kind (str):   Specifies the kind of interpolation. For more info see
                            `scipy.interpolate.interp1d`.

    Returns:
        proj_kernel (function):
                        A function of R (cylindrical radius / 'impact parameter')
                        of the projected (integrated along one coordinate) kernel.
    '''
    from scipy.interpolate import interp1d
    from scipy.integrate import quad

    kernel = kernels[kernel]
    Rs = np.arange(N+1,dtype=float) / N
    table = []
    for R in Rs:
        zmax = np.sqrt(1.0-R**2)
        w, err = quad(lambda z: kernel(np.sqrt(R**2+z**2)), -zmax, zmax)
        if err > 1e-6:
            raise RuntimeError('Encountered large error (%g) in integration!' %
                               err)
        table.append( w )
    return interp1d(Rs, table, copy=False, bounds_error=False, fill_value=0.0,
                    kind=inter_kind)

def integrate_kernel(kernel, N=100, inter_kind='quadratic'):
    '''
    Create an interpolation function for the 3D integrated kernel.

    The kernel w gets integrated as:

      /\
      |            2
      |  dx  4 pi x  w(x)
      |
    \/ [0,r]

    and this function (of r) gets returned.

    Args:
        kernel (str):       The name of the kernel to integrate (as in
                            kernels.kernel).
        N (int):            The number of points to use for the interpolation.
        inter_kind (str):   Specifies the kind of interpolation. For more info see
                            `scipy.interpolate.interp1d`.

    Returns:
        integ_kernel (function):    A function of r (radius) of the integrated
                                    kernel.
    '''
    from scipy.interpolate import interp1d
    from scipy.integrate import quad

    kernel = kernels[kernel]
    rs = np.linspace(0,1)
    table = []
    for r in rs:
        I, err = quad(lambda x: 4.*np.pi*x**2 * kernel(x), 0.0, r,
                      epsabs=1e-10, epsrel=1e-10)
        if err > 1e-6:
            raise RuntimeError('Encountered large error (%g) in integration!' %
                               err)
        table.append( I )
    return interp1d(rs, table, copy=False, bounds_error=False, fill_value=1.0,
                    kind=inter_kind)

#old version: needs to use 'intergrate_kernel' now (cache it?!)...
@static_vars( integs={} )
def rand_kernel_r(kernel=None):
    '''
    Get a random radius, with a PDF (propability density function) of the
    specified 3D kernel.

    Not the radii themselves have the PDF of the kernel, but if one would multiply
    them with vectors of length one and random direction in 3D, the resulting
    points in 3D space would have the PDF of the (3D-)kernel.

    Args:
        kernel (str):   The name of the kernel to use. (By default use the kernel
                        defined in `gadget.cfg`.)

    Returns:
        r (float):      A random radius.
    '''
    from scipy.optimize import brentq

    if kernel is None:
        from ..gadget import general
        kernel = general['kernel']

    # this caching gives a speed-up of a factor of an order of magnitude!
    if kernel not in rand_kernel_r.integs:
        rand_kernel_r.integs[kernel] = integrate_kernel(kernel)
    F = rand_kernel_r.integs[kernel]
    q = np.random.random()
    return brentq(lambda x: F(x)-q, 0,1)



'''
This module handles basic Lambda-CDM Friedman-Lemaitre-Robertson-Walter
cosmologies.

Example:
    >>> from ..environment import module_dir
    >>> cosmo = WMAP7()
    >>> if abs(cosmo.universe_age() - '13.77 Gyr') > '1e-2 Gyr':
    ...     print(cosmo.universe_age())
    >>> if abs(cosmo.cosmic_time(1.5) - '4.375 Gyr') > '1e-3 Gyr':
    ...     print(cosmo.cosmic_time(1.5))
    >>> z = cosmo.lookback_time_2z('5 Gyr') # reshift before 5 Gyr
    >>> assert abs(z - 0.4915) < 1e-4
    >>> assert abs(z2a(z) - 0.6705) < 1e-4  # and corresponding scale factor
'''
__all__ = ['a2z', 'z2a', 'Planck2013', 'WMAP7', 'FLRWCosmo']

import numpy as np
from ..units import *
from .quantities import *


def a2z(a):
    """
    Convert a scalefactor to redshift.

    Limited to z=1e15 (i.e. a2z(0)=1e15) in order to make the numerics more
    stable.

    Example:
        >>> a2z(1)
        0.0
        >>> round(a2z(0.1234), 4)
        7.1037
        >>> round(a2z(z2a(1.234)), 10)
        1.234
        >>> round( np.log10( a2z(0) ), 3 )
        15.0
    """
    return (1.0 - a) / (a + 1e-15)


def z2a(z):
    """
    Convert a scalefactor to redshift.

    Example:
        >>> z2a(0)
        1.0
        >>> z2a(12.345)
        0.07493443237167478
        >>> round(z2a(a2z(0.363)), 10)
        0.363
    """
    return 1.0 / (1.0 + z)


def _rho_crit(H, G):
    '''
    Critical density of the universe as inferred from the Hubble 'constant' H and
    the gravitational constant G.

    It simply executes:

                     2
                  3 H
    rho      =  --------
       crit      8 pi G

    (No care is taken of units!)

    Example:
        >>> _rho_crit(H=2.2815e-18, G=6.67384e-8)
        9.309932895584995e-30
    '''
    return 3.0 * H ** 2 / (8.0 * np.pi * G)


def Planck2013():
    '''
    Create a FLRW cosmology with the Planck 2013 results.

    Parameters are taken from Table 9 of the Planck Collaboration results of 2013.

    Returns:
        FLRWCosmo(h_0=0.6777, Omega_Lambda=0.683, Omega_m=0.307,
                  Omega_b=0.04825, sigma_8=0.8288, n_s=0.9611)

    Example:
        >>> planck = Planck2013()
        >>> planck.universe_age()
        UnitArr(13.786917087815011, units="Gyr")
        >>> planck.f_c
        0.15716612377850164
    '''
    return FLRWCosmo(h_0=0.6777, Omega_Lambda=0.683, Omega_m=0.307,
                     Omega_b=0.04825, sigma_8=0.8288, n_s=0.9611)


def WMAP7():
    '''
    Create a FLRW cosmology with the parameters of the 7-year Wilkinson Microwave
    Anisotropy Probe.

    Returns:
        FLRWCosmo(h_0=0.704, Omega_Lambda=0.728, Omega_m=0.272, Omega_b=0.0455,
                  sigma_8=810, n_s=0.967)

    Example:
        >>> wmap = WMAP7()
        >>> wmap.universe_age()
        UnitArr(13.768970244523558, units="Gyr")
        >>> wmap.f_c
        0.16727941176470587
    '''
    return FLRWCosmo(h_0=0.704, Omega_Lambda=0.728, Omega_m=0.272, Omega_b=0.0455,
                     sigma_8=0.810, n_s=0.967)


class FLRWCosmo(object):
    '''
    A Friedman-Lemaitre-Robertson-Walker cosmology that can calculate various
    cosmological values.

    Args:
        h_0 (float):            The Hubble parameter h (as in
                                H = 100 h km s**-1 Mpc**-1) at z=0.
        Omega_Lambda (float):   The density parameter of dark energy at z=0.
        Omega_m (float):        The density parameter of all (dark + baryonic)
                                matter at z=0.
        Omega_b (float):        The density parameter of baryonic matter at z=0.
                                Defaults to 0.16*Omega_m.
        sigma_8 (float):        Amplitude of the (linear) power spectrum at a
                                scale of 8 h**-1 Mph. (Not used yet.)
        n_s (float):            Scalar spectral index. (Not used yet.)

    Note:
        Radiation densities are neglected throughout.

    Example:
        >>> cosmo = FLRWCosmo(h_0=0.7, Omega_Lambda=0.7, Omega_m=0.3,
        ...                   Omega_b=0.05, sigma_8=0.8, n_s=0.96)
        >>> cosmo.Omega_tot
        1.0
        >>> cosmo.is_flat()
        True
        >>> assert abs(cosmo.h(1) - 1.2325) < 1e-4
        >>> assert abs(cosmo.f_c - 0.166667) < 1e-4
        >>> assert abs(cosmo.universe_age() - '13.48 Gyr') < '0.02 Gyr'
        >>> assert abs(cosmo.cosmic_time(2.0) - '3.229 Gyr') < '0.01 Gyr'
        >>> assert abs(cosmo.rho_crit()-'136.0 Msol/kpc**3') < '1. Msol/kpc**3'
        >>> assert abs(cosmo.rho_crit(10)-'5.44e4 Msol/kpc**3') < '1e2 Msol/kpc**3'
        >>> assert abs(cosmo.lookback_time_2z('5 Gyr') - 0.49383) < 1e-4
        >>> cosmo.H()
        UnitArr(70.0, units="Mpc**-1 km s**-1")
        >>> assert abs(cosmo.H(z=1.0) - '123.25 km/s/Mpc') < '0.1 km/s/Mpc'
        >>> z = 1.234
        >>> assert abs(cosmo.comoving_distance(z) - '3.839 Gpc') < '1e-3 Gpc'
        >>> assert abs(cosmo.trans_comoving_distance(z) - '3.839 Gpc') < '1e-3 Gpc'
        >>> assert abs(cosmo.angular_diameter_distance(z) - '1.719 Gpc') < '1e-3 Gpc'
        >>> assert abs(cosmo.luminosity_distance(z) - '8.577 Gpc') < '1e-3 Gpc'
        >>> assert (abs(cosmo.apparent_angle('1 kpc', z) - '0.1200 arcsec')
        ...         < '1e-4 arcsec')
        >>> assert (abs(cosmo.angle_to_length('1 arcsec', z) - '8.332 kpc')
        ...         < '1e-3 kpc')
        >>> assert abs(cosmo.angle_to_length(1e-6, z) - '1.719 kpc') < '1e-3 kpc'
        >>> assert (abs(cosmo.angle_to_length(cosmo.apparent_angle('1 kpc',z), z) -
        ...         '1 kpc') < '1e-5 kpc')
    '''

    def __init__(self, h_0, Omega_Lambda, Omega_m, Omega_b=None, sigma_8=None,
                 n_s=None):
        if sigma_8:
            assert sigma_8 > 0
        self._h_0 = h_0
        self._Omega_Lambda = Omega_Lambda
        self._Omega_m = Omega_m
        self._Omega_b = 0.16 * Omega_m if Omega_b is None else Omega_b
        self._sigma_8 = sigma_8
        self._n_s = n_s  # not used by now

    def __repr__(self):
        return 'FLRWCosmo(h_0=%.4g, ' % self._h_0 + \
               'O_Lambda=%.4g, ' % self._Omega_Lambda + \
               'O_m=%.4g, ' % self._Omega_m + \
               'O_b=%.4g, ' % self._Omega_b + \
               'sigma_8=%s, ' % ('None' if self._sigma_8 is None \
                                     else '%.4g' % self._sigma_8) + \
               'n_s=%s)' % ('None' if self._n_s is None else '%.4g' % self._n_s)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        return (self._h_0 == other._h_0 and
                self._Omega_Lambda == other._Omega_Lambda and
                self._Omega_m == self._Omega_m and
                self._Omega_b == self._Omega_b and
                self._sigma_8 == self._sigma_8 and
                self._n_s == self._n_s)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __cmp__(self, other):
        return None

    def copy(self):
        '''Create a copy of the class instance.'''
        copy = object.__new__(FLRWCosmo)
        copy.__dict__ = self.__dict__.copy()
        return copy

    @property
    def h_0(self):
        '''Dimensionless Hubble parameter at z=0.'''
        return self._h_0

    @property
    def Omega_Lambda(self):
        '''Dark energy density parameter (at z=0).'''
        return self._Omega_Lambda

    @property
    def Omega_m(self):
        '''Matter (total) density parameter (at z=0).'''
        return self._Omega_m

    @property
    def Omega_b(self):
        '''Baryon density parameter (at z=0).'''
        return self._Omega_b

    @property
    def Omega_k(self):
        '''Curvature density parameter (at z=0).'''
        return 1.0 - self._Omega_m - self._Omega_Lambda

    @property
    def Omega_tot(self):
        '''Density (total) parameter (at z=0).'''
        return self._Omega_m + self._Omega_Lambda

    @property
    def f_c(self):
        '''Cosmic baryon fraction (at z=0).'''
        return self._Omega_b / self._Omega_m

    @property
    def sigma_8(self):
        '''Density (total) parameter (at z=0).'''
        return self._sigma_8

    @property
    def n_s(self):
        '''Density (total) parameter (at z=0).'''
        return self._n_s

    @h_0.setter
    def h_0(self, value):
        self._h_0 = value

    @Omega_Lambda.setter
    def Omega_Lambda(self, value):
        self._Omega_Lambda = value

    @Omega_m.setter
    def Omega_m(self, value):
        assert value >= 0
        self._Omega_m = value

    @Omega_b.setter
    def Omega_b(self, value):
        assert (value >= 0 and value <= self._Omega_m)
        self._Omega_b = value

    @sigma_8.setter
    def sigma_8(self, value):
        assert sigma_8 >= 0
        self._sigma_8 = value

    @n_s.setter
    def n_s(self, value):
        self._n_s = value

    def is_flat(self, tol=0.0):
        '''Test whether the cosmology is a flat one within the tolerance 'tol'
        (abs(self.Omega_k) <= tol).'''
        return abs(self.Omega_k) <= tol

    def _E(self, z=0.0):
        '''Helper function for e.g. FLRWCosmo.h() and
        FLRWCosmo.lookback_time().'''
        if z < -1:
            raise ValueError('Redshift cannot be smaller -1.')
        return np.sqrt(self._Omega_m * (1.0 + z) ** 3
                       + self.Omega_k * (1.0 + z) ** 2
                       + self._Omega_Lambda)

    def h(self, z=0.0):
        '''Calculate the dimensionless Hubble parameter at redshift z.'''
        return self._h_0 * self._E(z)

    def H(self, z=0.0):
        '''Calculate the Hubble parameter at redshift z.'''
        return UnitArr(100.0 * self.h(z=z), 'km/s / Mpc')

    def rho_crit(self, z=0.0):
        '''
        Calculate the critical density at redshift z.

        Note:
            The function uses G, from the quantities module.

        Args:
            z (float):  The redshift to calculate rho_crit for.

        Returns:
            rho_crit (UnitArr): The critical density (by default in units of
                                'Msol / kpc**3').
        '''
        from ..snapshot.snapshot import Snapshot
        rho_crit = 3.0 * self.H(z) ** 2 / (8.0 * np.pi * G)
        rho_crit.convert_to('Msol / kpc**3')
        return rho_crit

    def lookback_time_in_Gyr(self, z):
        '''
        Lookback time (as float) for an object with redshift z in Gyr.

        Note:
            This function is a few times faster than FLRWCosmo.lookback_time(z)
            and uses 1 pc = 3.08567758149e+16 m and 1 yr = 365 d = 31536000 s.
        '''
        # this import is suspended until here, because it takes some time
        from scipy.integrate import quad

        if z <= -1:
            raise ValueError('Redshift cannot be smaller -1.')
        t, err = quad(lambda zz: 1.0 / ((1.0 + zz) * self._E(zz)), 0.0, z)
        # divide by H to get t in units of 1/(km s**-1 Mpc**-1)
        t /= 100.0 * self._h_0
        # 978.461942380137 Gyr = 1 / (km*s**-1 * Mpc**-1)
        return t * 978.461942380137

    def lookback_time(self, z):
        '''
        Lookback time for an object with redshift z.

        Note:
            Internally this function uses self.lookback_time_in_Gyr.

        Args:
            z (float):  The redshift to calculate the lookback time for.

        Returns:
            lookback_time (UnitArr):    Lookback time.
        '''
        return UnitArr(self.lookback_time_in_Gyr(z), 'Gyr')

    def universe_age(self):
        '''
        The present day age of the universe.

        Note:
            The function calls FLRWCosmo.lookback_time(z=oo) in the back.
        '''
        return self.lookback_time(z=np.inf)

    def cosmic_time(self, z):
        '''
        The cosmic time at a given redshift.

        Args:
            z (float):  The redshift to calculate the cosmic time for.
        '''
        return self.lookback_time(z=np.inf) - self.lookback_time(z=z)

    def lookback_time_2z(self, t):
        '''
        Convert a lookback time in given units to a redshift.

        Note:
            The algorithm only converges, if the result for the redshift is
            smaller than ~1e5 (about 550 years after the big bang in our universe)
            and t is positive.

        Args:
            t (UnitScalar): The cosmic time. (If a float is passed, it is
                            interpreted as Gyr.)
        '''
        # this import is suspended until here, because it takes some time
        from scipy.optimize import brentq

        t = UnitScalar(t, 'Gyr', dtype=float)

        if self.universe_age() < t:
            raise ValueError('Lookback time (%s) is larger than the age ' % t + \
                             'of the universe (%s). ' % self.universe_age() + \
                             'Cannot convert to redshift.')
        if t < 0:
            raise ValueError('Does not support negative lookback times.')
        t = float(t.in_units_of('Gyr'))
        return brentq(lambda z: self.lookback_time_in_Gyr(z) - t, -1e-10, 1e5,
                      maxiter=1000, disp=True)

    def comoving_distance(self, z, units='Gpc'):
        '''
        Comoving distance for an object with redshift z.

        Args:
            z (float):          The redshift to calculate the distance for.
            units (str, Unit):  The units to return the distance in.

        Returns:
            comoving_distance (UnitArr):    Comoving distance.
        '''
        from scipy.integrate import quad

        if z <= -1:
            raise ValueError('Redshift cannot be smaller -1.')
        d_c, err = quad(lambda zz: 1.0 / self._E(zz), 0.0, z)
        # divide by H to get d_c (com. dist. over speed of light) in units of
        # 1/(km s**-1 Mpc**-1)
        d_c /= 100.0 * self._h_0
        # 978.461942380137 Gyr = 1 / (km*s**-1 * Mpc**-1)
        d_c *= 978.461942380137
        return (c * UnitArr(d_c, 'Gyr')).in_units_of(units, subs={'a': 1.0})

    def trans_comoving_distance(self, z, units='Gpc'):
        '''
        The transverse comoving distance for an object with redshift z.

        Args and Returns as in comoving_distance.
        '''
        dc = self.comoving_distance(z=z, units=units)
        if self.Omega_k == 0.0:
            return dc

        dh = (c / self.H()).in_units_of(units)
        if self.Omega_k > 0.0:
            sqrt_Ok = np.sqrt(self.Omega_k)
            return dh / sqrt_Ok * np.sinh(sqrt_Ok * dc / dh)
        else:  # self.Omega_k < 0.0
            sqrt_Ok = np.sqrt(-self.Omega_k)
            return dh / sqrt_Ok * np.sin(sqrt_Ok * dc / dh)

    def angular_diameter_distance(self, z, units='Gpc'):
        '''
        The angular diameter distance for an object with redshift z.

        Args and Returns as in comoving_distance.
        '''
        return self.trans_comoving_distance(z, units=units) / (1.0 + z)

    def luminosity_distance(self, z, units='Gpc'):
        '''
        The luminosity distance for an object with redshift z.

        Args and Returns as in comoving_distance.
        '''
        return (1.0 + z) * self.trans_comoving_distance(z, units=units)

    def apparent_angle(self, l, z, units='arcsec'):
        '''
        Convert a physical extent of some object at redshift z to its apparent
        angle.

        Args:
            l (UnitScalar):     The physical length of the object.
            z (float):          The redshift of the object.
            units (str, Unit):  The units to return the angle in (has to be an
                                angular unit, cannot be unitless!).

        Returns:
            theta (UnitArr):    The object's apparent angle.
        '''
        l = UnitScalar(l)
        dA = self.angular_diameter_distance(z, units=l.units)
        theta = UnitArr(float(l / dA), 'rad')
        return theta.in_units_of(units)

    def angle_to_length(self, theta, z, units='kpc'):
        '''
        Convert an apparent angle of some object at redshift z into its physical
        extent.

        Args:
            theta (UnitScalar, float):
                                The apparent angle of the object. If a float is
                                passed, it is interpreted as radiants.
            z (float):          The redshift of the object.
            units (str, Unit):  The units to return the extent in.

        Returns:
            ext (UnitArr):      The object's physical extent.
        '''
        theta = UnitScalar(theta, 'rad')
        dA = self.angular_diameter_distance(z, units=units)
        return float(theta) * dA




'''
All about kernels

Also doctest other parts of this sub-module:
    >>> import doctest
    >>> doctest.testmod(definitions)
    TestResults(failed=0, attempted=5)
    >>> doctest.testmod(integral)
    TestResults(failed=0, attempted=9)
'''

from .definitions import *
from .integral import *




'''
A collection of analysis functions to create profiles.

Example:
    >>> from ..environment import module_dir
    >>> from ..snapshot import Snapshot
    >>> s = Snapshot(module_dir+'snaps/snap_M1196_4x_320', physical=True)
    >>> center = UnitArr([33816.9, 34601.1, 32681.0], s['pos'].units)
    load block pos... done.
    >>> radially_binned(s.stars, 'Fe/mass', av='mass', center=center,
    ...                 r_edges=np.linspace(0,15,30))
    load block Z... done.
    derive block elements... done.
    derive block Fe... done.
    load block mass... done.
    UnitArr([0.00080558, 0.00065764, 0.00053465, 0.00049411, 0.00046332,
             0.00043347, 0.00043268, 0.00042417, 0.00040976, 0.00039034,
             0.00039317, 0.00037786, 0.00036477, 0.00035503, 0.00035541,
             0.00034816, 0.00030562, 0.00031812, 0.00030737, 0.00032145,
             0.00029885, 0.00028331, 0.00028867, 0.00029243, 0.00029605,
             0.00028672, 0.00029282, 0.00027797, 0.00026345], dtype=float32)
    >>> profile_dens(s, 'mass', center=center,
    ...     r_edges=[0]+list(np.logspace(-1,2,30))) # doctest:+ELLIPSIS
    UnitArr([4.94145445e+09, 2.23835198e+09, 1.00754934e+09, 1.45437910e+09,
             1.62901107e+09, 2.11259075e+09, 1.67041258e+09, 1.19817515e+09,
             8.61809349e+08, 6.60993375e+08, 4.51407479e+08, 3.60605928e+08,
             2.54610455e+08, 1.74468352e+08, 1.23857127e+08, 8.19490250e+07,
             5.28403138e+07, 3.39792774e+07, 2.04022431e+07, 1.18351359e+07,
             6.89966816e+06, 4.16610461e+06, 2.27211531e+06, 1.30861066e+06,
             7.09065373e+05, 3.94554812e+05, 2.16913872e+05, 1.50912836e+05,
             9.56563740e+04, 5.45304130e+04], units="Msol kpc**-3")
    >>> rho0, Rs, err = NFW_fit(s, center, Rmax='300 kpc')  # R200 ~ 178 kpc
    >>> if (abs(rho0 - '4.46e7 Msol/kpc**3') > '0.03e7 Msol/kpc**3'
    ...             or abs(Rs - '9.68 kpc') > '0.02 kpc'):
    ...     print(rho0, Rs)
    >>> if abs(err - 0.13) > 0.01:
    ...     print(err)
    >>> smassprof1 = radially_binned(s.stars, 'mass', center=center)
    >>> from ..transformation import Translation
    >>> Translation(-center).apply(s)
    apply Translation to "pos" of "snap_M1196_4x_320"... done.
    >>> smassprof2 = radially_binned(s.stars, 'mass')
    derive block r... done.
    >>> assert (np.abs(smassprof1-smassprof2)/smassprof1).max() < 1e-3
'''
__all__ = ['radially_binned', 'profile_dens', 'NFW', 'NFW_fit']

import numpy as np
from ..units import *
from ..utils import dist

def radially_binned(s, qty, av=None, r_edges=None, proj=None, center=None):
    '''
    Bin a quantity radially.

    Args:
        s (Snap):               The snapshot to use.
        qty (str, Unitqty):     The quantity to do the binning of.
        av (str, Unitqty):      The quantity to average over. Otherwise as 'qty'.
        r_edges (array-like):   The edges of the radial bins.
                                Default: 0 - 20 kpc in 50 bins
        proj (int):             If None bin spherically, otherwise bin projected
                                along this coordinate: (0:'x', 1:'y', 2:'z').
        center (array-like):    The center with respect to which the radii are
                                calculated.

    Returns:
        Q (UnitArr):            The binned quantity.
    '''
    if isinstance(qty, str):
        qty = s.get(qty)
    else:
        qty = UnitQty(qty)
    if av is not None:
        if isinstance(av, str):
            av = s.get(av)
        else:
            av = UnitQty(av)

    if center is None:
        center = [0,0,0] if proj is None else [0,0]
    center = UnitQty(center, s['pos'].units, subs=s)

    if proj is None:
        r = s['r'] if np.all(center==0) else dist(s['pos'],center)
    else:
        if proj not in list(range(3)):
            raise ValueError('Have to project along 0, 1, or 2!')
        if np.all(center==0) and proj==2:
            r = s['rcyl']
        else:
            axes = tuple( set([0,1,2]) - set([proj]) )
            r = dist(s['pos'][:,axes],center)
    r_ind = r.argsort()

    if r_edges is None:
        r_edges = UnitArr(np.linspace(0,20,51), 'kpc')
    r_edges = UnitQty(r_edges, s['pos'].units, subs=s)

    # r_edges as indices in r_ind
    ind_edges = np.array( [np.abs(r[r_ind]-rr).argmin() for rr in r_edges] )

    sorted_qty = qty[r_ind]
    if av is not None:
        sorted_av = av[r_ind]
        sorted_qty = sorted_qty * sorted_av
    Q = []
    for i in range(1, len(r_edges)):
        val = sorted_qty[ind_edges[i-1]:ind_edges[i]].sum()
        if av is not None:
            val /= sorted_av[ind_edges[i-1]:ind_edges[i]].sum()
        Q.append( val )

    Q = UnitArr(Q,qty.units)

    if av is not None:
        Q[np.isnan(Q)] = 0.0

    return Q

def profile_dens(s, qty, av=None, r_edges=None, proj=None, center=None):
    '''
    Create a radial profile for the density of a quantity.

    It calls radially_binned and then devides the bins by their volume / area.

    Args:
        See radially_binned.

    Returns:
        profile (UnitArr):      The profile in units of the quantity per
                                volume/area.
    '''
    if r_edges is None:
        r_edges = UnitArr(np.linspace(0,20,51), 'kpc')
    r_edges = UnitQty(r_edges, s['pos'].units, subs=s)

    Q = radially_binned(s, qty, av=av, r_edges=r_edges, proj=proj, center=center)
    if proj is None:
        V = (4.0/3.0*np.pi)*r_edges**3
    else:
        V = np.pi*r_edges**2
    V = V[1:] - V[:-1]
    return Q/V

def NFW(r, rho0, Rs):
    '''
    The function for a Navarro-Frenk-White (NFW) profile.

    Note:
        If both r and Rs are integers, an integer division is performed there and
        the result is probably wrong!

    Returns:
        rho = rho0 / ( r/Rs * (1.0 + r/Rs)**2 )

    Examples:
        >>> NFW(0.1, 42.0, 1.234)
        443.4892634342499
        >>> NFW(12.34, 2.1, 12.34)
        0.525
        >>> NFW(123.4, 1.0, 43.21)
        0.0235524185442844
    '''
    r_norm = r/Rs
    return rho0 / ( r_norm * (1.0 + r_norm)**2 )

def _NFW_log10(lr, lrho0, lRs):
    '''
    The logarithmic version of the NFW profile for fitting.

    Doctests:
        >>> assert abs(_NFW_log10(-1,1.23,0.12)
        ...             - np.log10(NFW(0.1,10**1.23,10**0.12))) < 1e-5
        >>> assert abs(_NFW_log10(2.1,12.3,1.2)
        ...             - np.log10(NFW(10.0**2.1,10**12.3,10**1.2))) < 1e-5
    '''
    r_norm = 10.0**lr / 10.0**lRs
    return lrho0 - np.log10( r_norm * (1.0 + r_norm)**2 )
def NFW_fit(s, center=None, Rmin=None, Rmax=None, Nfit=100, r_edges=None):
    '''
    Do a Navarro-Frenk-White (NFW) profile fit.

    First a logarithmic density profile is created with logarithmic radial bins.
    Then a least-square fit of a NFW profile (cf. NFW function) is done to this
    profile.

    Args:
        s (Snap):               The (sub-)snapshot to use (all particles are
                                used).
        center (array-like):    The center to take the profiles from.
        Rmin (float, UnitArr):  The minimal radius for the profile. Default 1 kpc.
        Rmax (float, UnitArr):
                                The radius to fit up to. (Should be close to
                                virial radius). Default: 100 kpc.
        Nfit (int):             The number of points of the profile to fit.
        r_edges (array-like):   The edges of the radial bins. If given Rmin, Rmax,
                                and Nfit are ignored.

    Returns:
        rho0 (UnitArr): The logarithm of the density parameter of the fit.
        Rs (UnitArr):   The scale radius.
        err (array):    The standard deviation error in the *logarithmic* fit.
    '''
    from scipy.optimize import curve_fit

    if r_edges is None:
        if Rmin is None: Rmin = '1 kpc'
        if Rmax is None: Rmax = '100 kpc'
        Rmin = float(UnitScalar(Rmin,s['pos'].units,subs=s))
        Rmax = float(UnitScalar(Rmax,s['pos'].units,subs=s))
        r_edges = np.array([0]+list(np.logspace(np.log10(Rmin),np.log10(Rmax),Nfit)))
    else:
        r_edges = UnitQty(r_edges, s['pos'].units, subs=s)
    # get something like the midpoints
    r = np.array( [r_edges[1]] + list((r_edges[2:]+r_edges[1:-1])/2.) )
    rho = profile_dens(s, 'mass', center=center, r_edges=r_edges, proj=None)

    # logarithmic fitting
    log_r   = np.log10(r).view(np.ndarray)
    log_rho = np.log10(rho).view(np.ndarray)
    log_Rs      = log_r[-1] - 1.0   # divide (a bit less than) Rmax by 10
    log_rho0    = log_rho[np.abs(log_r-log_Rs).argmin()]
    (log_rho0, log_Rs), cov = curve_fit(_NFW_log10, log_r, log_rho,
                                        p0=(log_rho0,log_Rs))

    # fit quality
    err = (_NFW_log10(log_r, log_rho0, log_Rs) - log_rho).std()

    return (UnitArr(10.0**log_rho0, rho.units),
            UnitArr(10.0**log_Rs, s['pos'].units),
            err)




'''
Module for plotting.

Also doctest other parts of this sub-module:
    >>> import doctest
    >>> doctest.testmod(general)
    TestResults(failed=0, attempted=1)
'''

from .colormaps import *
from .general import *
from .maps import *
from .profiles import *

def show_FoF_groups(s, groups, colors='rand', plot_center=True, plot_parts=False,
                    plot_hull=True, pointsize=20, alpha=0.02, linewidth=1,
                    linestyle='solid', **plotargs):
    '''
    Plot the image of the (sub-)snapshot `s` and overplot the convex hulls of the
    (projected) FoF-groups.

    Args:
        s (Snap):           The (sub-)snapshot from which the FoF-groups where
                            drawn.
        groups (list):      A list of the FoF groups as instances of ???.
        colors:             The line color. Can be:
                            * 'rand':       Choose a new random color for each
                                            FoF-group.
                            * str/3-tuple:  Define one color for all FoF-groups.
                            * iterable:     Individual colors for the FoF-groups.
        plot_center (str):  Whether to plot the centers of the groups and which
                            one (as attribute name of a group). If True, it will
                            be the center of mass.
        plot_parts (bool):  Whether to do a scatter plot of the particles of the
                            FoF-groups. The are transparent with an alpha as
                            specified.
        plot_hull (bool):   Whether to plot the convex hull.
        pointsize (float):  The side of the points for the scatter plot and the
                            centers markers ('plot_parts' and 'plot_center').
        alpha (float):      The alpha value for the scatter plot of the FoF-group.
        linewidth (float):  The linewidth for the convex hulls (constant for all
                            FoF-groups).
        linestyle (str):    The linestyle for the convex hulls (constant for all
                            FoF-groups).
        **plotargs:         Further arguments are passed to `image`.

    Returns:
        fig, ax, im, cbar:  As `image` does.
    '''
    import numpy as np
    import matplotlib as mpl
    from scipy.spatial import ConvexHull

    xaxis = plotargs.get('xaxis', 0)
    yaxis = plotargs.get('yaxis', 1)
    if not isinstance(colors, str):
        colors = np.asarray(colors)

    fig, ax, im, cbar = image(s, **plotargs)
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()

    for i, group in enumerate(groups):
        if colors == 'rand':
            color = mpl.colors.hsv_to_rgb( [np.random.rand(), 1, 1] )
        elif isinstance(colors,str) or colors.dtype in (str,str) \
                or colors.shape==(3,):
            color = colors
        else:
            color = colors[i]

        pos2D = s[group]['pos'][:,(xaxis,yaxis)]
        if plot_hull:
            hull = ConvexHull(pos2D)
            ax.fill(pos2D[hull.vertices,0], pos2D[hull.vertices,1], fill=False,
                    edgecolor=color, linestyle=linestyle, linewidth=linewidth)
        if plot_parts:
            ax.scatter(pos2D[:,0], pos2D[:,1], s=pointsize,
                       edgecolor='none', facecolor=color, alpha=alpha)
        if plot_center:
            try:
                center = getattr(group, plot_center)
            except:
                center = group.com
            c = mpl.colors.rgb_to_hsv(color)
            c = mpl.colors.hsv_to_rgb( [c[0], c[1]*0.6, c[2]*0.8] )
            ax.scatter(center[xaxis], center[yaxis],
                       marker='x', s=1.5*pointsize, color=c)

    ax.set_xlim(xlim)
    ax.set_ylim(ylim)

    return fig, ax, im, cbar




'''
Read and interpolate Cloudy tables for ionisation states as a function of density
and temperature for a given redshift.
'''
__all__ = ['config_ion_table', 'IonisationTable']

from .. import environment
from ..units import Unit, UnitQty
from .. import gadget
from .. import physics
import numpy as np
import copy
import re
import os

def config_ion_table(redshift):
    '''
    Load & interpolate an IonisationTable for the given redshift as defined in the
    config file `derived.cfg`.

    Args:
        redshift (float):   The redshift to load/interpolate the table(s) for.

    Returns:
        iontbl (IonisationTable):   The ionisation table class.
    '''
    from ..snapshot.derived import iontable
    if iontable['tabledir'] is None:
        raise RuntimeError('No Cloudy table directory defined in `derived.cfg`!')

    #for old style Oppenheimer tables:
    nH_vals = iontable['nH_vals']
    nH_vals = np.linspace(float(nH_vals[0]),
                          float(nH_vals[0] + nH_vals[1]*(nH_vals[2]-1)),
                          nH_vals[2])
    T_vals = iontable['T_vals']
    T_vals = np.linspace(float(T_vals[0]),
                         float(T_vals[0] + T_vals[1]*(T_vals[2]-1)),
                         T_vals[2])

    iontbl = IonisationTable(redshift,
                             tabledir=iontable['tabledir'],
                             table_pattern=iontable['pattern'],
                             flux_factor=iontable['flux_factor'],
                             style=iontable['style'],
                             nH=nH_vals,
                             T=T_vals,
                             ions=iontable.get('ions',[]))
    return iontbl

class IonisationTable(object):
    '''
    Read in a Cloudy table for fractions of ions and prepare it for interpolation.

    Note:
        For debugging purposes, there is a method to plot the tables read in:
        `show_table`.

    Args:
        redshift (float):       The redshift to create the table for. The two
                                tables inbetween which the redshift is are loaded
                                and linearly interpolated.
        tabledir (str):         The directory where all the tables are located.
        table_pattern (str):    The pattern of the table file names. It is their
                                filename with '<z>' where the redshift is encoded.
                                This code is 10*z rounded to the next integer.
        flux_factor (float):    Adjust the UVB by this factor (assume a optically
                                thin limit and scale down the densities during the
                                lookup by this factor).
                                (Note: `sigHI` used in the Rahmati fitting formula
                                needed for an approximation of self-shielding is
                                not adjusted!)
        style (str):            Either 'Oppenheimer old' or 'Oppenheimer new'.
        nH (array-like):        The Hydrogen number densities of the grid points
                                in units of log10(cm^-3).
        T (array-like):         The temperatures for the grid points in units of
                                log10(K).
        ions (iterable):        The names of the ions in the table.
    '''

    def __init__(self, redshift, tabledir,
                 table_pattern='lt<z>f100_i45', flux_factor=1.0,
                 style='Oppenheimer new', nH=None, T=None, ions=None):
        if style not in ['Oppenheimer old', 'Oppenheimer new']:
            raise ValueError('Unknown style "%s"!' % style)
        self._style = style

        if style == 'Oppenheimer old':
            nH = np.asarray(nH)
            T = np.asarray(T)
            if nH.ndim!=1 or T.ndim!=1:
                raise ValueError('`nH` and `T` need 1-dim., see docs!')
            self._nH_vals = nH.copy()
            self._T_vals = T.copy()
            self._ions = list(ions)
        else:
            self._nH_vals = None
            self._T_vals = None
            self._ions = None
        if table_pattern.count('<z>') != 1:
            raise ValueError('The table pattern has to have exactly one ' +
                             'occurrence of "<z>"!')

        self._redshift = float(redshift)
        self._tabledir = tabledir
        self._pattern = table_pattern
        self._table = self._get_table()
        self.flux_factor = flux_factor

    @property
    def redshift(self):
        return self._redshift
    @property
    def tabledir(self):
        return self._tabledir
    @property
    def table_pattern(self):
        return self._pattern
    @property
    def nH_vals(self):
        return self._nH_vals
    @property
    def T_vals(self):
        return self._T_vals
    @property
    def ions(self):
        return self._ions
    @property
    def data(self):
        return self._table
    @property
    def flux_factor(self):
        return self._flux_factor
    @flux_factor.setter
    def flux_factor(self, value):
        if float(value) <= 0.0:
            raise ValueError('The flux factor needs to be positive!')
        self._flux_factor = float(value)

    def interp_snap(self, ion, s, nH='nH', T='temp', selfshield=False,
                    warn_outofbounds=True):
        '''
        Convenience function: calling `interp_parts(ion, s.get(nH), s.get(T),
        selfshield, subs=s, warn_outofbounds=warn_outofbounds)`.
        '''
        if abs(s.redshift - self._redshift) > 0.01:
            import sys
            print('WARNING: the snapshot\'s redshift ' + \
                                 '(%.3f) does not match the ' % s.redshift + \
                                 'table\'s redshift (%.3f)!' % self._redshift, file=sys.stderr)
        return self.interp_parts(ion, s.get(nH), s.get(T), selfshield, subs=s,
                                 warn_outofbounds=warn_outofbounds)

    def interp_parts(self, ion, nH, T, selfshield=True, UVB=gadget.general['UVB'],
                     z=None, fbaryon=None, subs=None, warn_outofbounds=True):
        '''
        Given nH and temperatures for a list of particles, get the fractions for
        a given ion by a bilinear interpolation of the table for each particle.

        If values out of the table bounds are encountered, the corresponding
        values will be the nearest table values.

        Args:
            ion (int,str):      The index of name of the ion in question.
            nH (UnitQty):       The Hydrogen number densities of the particles (if
                                given unitless, interpreted as in cm^-3).
            T (UnitQty):        The temperatures of the particles (if given
                                unitless, interpreted as in K).
            selfshield (bool):  Whether to account for self-shielding by assuming
                                the attenuation of the UVB as experienced by HI
                                using the Rahmati+ (2013) prescription (formula
                                (14) of the paper). Note that this is just a rough
                                approximation!
            UVB (str):          The UVB background used for the self-shielding. If
                                no self-shielding is used, the argument is ignored.
            z (float):          The redshift used for the self-shielding. If None
                                and subs is a (sub-)snapshot, it is taken from the
                                snapshot. If no self-shielding is used, the
                                argument is ignored.
            fbaryon (float):    The cosmic baryon density used for the
                                self-shielding. If None and subs is a
                                (sub-)snapshot, it is taken from the snapshot's
                                cosmology. If no self-shielding is used, the
                                argument is ignored.
            subs (dict, Snap):  Used for substitutions that might be neccessary
                                when converting units (for more info see
                                `UnitArr.in_units_of`).
            warn_outofbounds (bool):
                                Whether to print out a warning if particles are
                                out of the bounds of the ionisation tables.

        Returns:
            f (np.ndarray):     The fractions of the ion for each particle as read
                                / interpolated from the table.
        '''
        if len(nH) != len(T):
            raise ValueError('Lengths of nH (%d) and T (%d) do not match!' % (
                                len(nH),len(T)))
        if isinstance(ion,str):
            ion = self._ions.index(str(ion))
        if ion<0 or self._table.shape[-1]<=ion:
            raise ValueError('Ion index (%d) out of bounds!' % ion)
        # units of table
        nH = UnitQty(nH, 'cm**-3', subs=subs)
        T  = UnitQty(T , 'K',      subs=subs)

        if selfshield:
            # with Rahmati+ (2013) description
            from ..snapshot.snapshot import Snapshot
            from ..cloudy import Rahmati_fGamma_HI
            if isinstance(subs,Snapshot):
                if z is None:
                    z = subs.redshift
                if fbaryon is None:
                    fbaryon = subs.cosmology.Omega_b / subs.cosmology.Omega_m
            else:
                if z is None:
                    raise ValueError('Redshift is needed for self-shielding!')
                if fbaryon is None:
                    # just some reasonable value
                    cosmo = physics.Planck2013()
                    fbaryon = cosmo.Omega_b / cosmo.Omega_m
                    import sys
                    print('WARNING: fbaryon was not given, using', fbaryon, file=sys.stderr)
            fG = Rahmati_fGamma_HI(z, nH, T, fbaryon, UVB=UVB,
                                   flux_factor=self.flux_factor, subs=subs)
            # limit the attenuation in very dense gas (assuming not all of the
            # particle mass actually is at these high densities)
            fG_limit = 1e-3
            fG[ fG < fG_limit ] = fG_limit
            # decreasing the radiation by a factor of fG corresponds to increasing
            # the density by the same factor fG:
            nH /= fG

        # assume the optical thin limit (as done for the self-shielding, too)
        if self._flux_factor != 1.0:
            nH /= self._flux_factor

        # look-up tables use log10-values
        nH = np.log10( nH )
        T  = np.log10( T  )

        # find lower indices
        N_nH = len(self._nH_vals)
        N_T  = len(self._T_vals)
        k_nH = -np.ones(len(nH), dtype=int)
        for k in range( N_nH ):
            k_nH[ self._nH_vals[k] < nH ] = k
        l_T = -np.ones(len(T), dtype=int)
        for l in range( N_T ):
            l_T[ self._T_vals[l] < T ] = l

        # do bilinear interpolation

        low_nH  = (k_nH<0)
        high_nH = (k_nH>=N_nH-1)
        low_T   = (l_T <0)
        high_T  = (l_T >=N_T -1)
        if warn_outofbounds:
            out_of_bounds = np.sum( low_nH | high_nH | low_T | high_T )
            if out_of_bounds:
                import sys
                from .. import utils
                print('WARNING: %s particles out of bounds!' % \
                        utils.nice_big_num_str(out_of_bounds), file=sys.stderr)
        # avoid invalid indices in the following (masking would slow down and be
        # not very readable):
        k_nH[ low_nH  ] = 0
        k_nH[ high_nH ] = N_nH-2
        l_T [ low_T   ] = 0
        l_T [ high_T  ] = N_T-2

        tbl = self._table[:,:,ion].ravel()
        f = np.empty(len(T), dtype=float)   # allocate memory

        f11 = tbl[  l_T    + (k_nH  ) * N_T]
        f21 = tbl[  l_T    + (k_nH+1) * N_T]
        f12 = tbl[ (l_T+1) + (k_nH  ) * N_T]
        f22 = tbl[ (l_T+1) + (k_nH+1) * N_T]
        # interpolate in nH
        nH1, nH2 = self._nH_vals[k_nH], self._nH_vals[k_nH+1]
        a = (nH - nH1) / (nH2 - nH1)
        f1 = (1.-a) * f11 + a * f21
        f2 = (1.-a) * f12 + a * f22
        f1[low_nH]  = f11[low_nH]
        f1[high_nH] = f21[high_nH]
        f2[low_nH]  = f12[low_nH]
        f2[high_nH] = f22[high_nH]
        # interpolate in T
        T1, T2 = self._T_vals[l_T], self._T_vals[l_T+1]
        a = (T - T1) / (T2 - T1)
        f = (1.-a) * f1 + a * f2
        f[low_T]  = f1[low_T]
        f[high_T] = f2[high_T]

        return f

    def _read_cloudy_table(self, filename):
        if self._style == 'Oppenheimer old':
            tbl = np.loadtxt(filename)
            if len(self._nH_vals) * len(self._T_vals) != tbl.shape[0]:
                raise ValueError("table size (%d) " % tbl.shape[0] +
                                 "does not match the grid given " +
                                 "(%dx%d)!" % (len(self._nH_vals),len(self._T_vals)))
            if len(self._ions) != tbl.shape[1]:
                raise ValueError("table column size (%d) " % tbl.shape[1] +
                                 "does not match the number of ions " +
                                 "(%d)!" % len(self._ions))
            tbl = tbl.reshape( (len(self._nH_vals), len(self._T_vals),
                                tbl.shape[1]) )
            redshift    = None
            nH_vals     = self._nH_vals
            T_vals      = self._T_vals
            short_ions  = self._ions
        elif self._style == 'Oppenheimer new':
            tbl = np.loadtxt(filename, skiprows=1)
            with open(filename, 'r') as f:
                head = f.readline().strip().strip('#')
            head = head.split()
            if len(head)-2!=tbl.shape[1] or head[-2]!='redshift=' or \
                    head[:2]!=['Hdens', 'Temp']:
                raise ValueError('Head of table "%s" of unexpected format!' % filename)

            redshift = float(head[-1])
            ions = head[2:-2]
            short_ions = []
            for ion in ions:
                i = [n for n,c in enumerate(ion) if c.isupper()][1]
                abb_el, state = ion[:i], ion[i:]
                # element name might be abbreviated, if ion name is to long otherwise
                el = None
                for full_el, short in physics.cooling.SHORT_ELEMENT_NAME.items():
                    if full_el[:len(abb_el)] == abb_el:
                        el = short
                        break
                if el is None:
                    raise RuntimeError('Short name of "%s[...]" not found!' % abb_el)
                short_ions.append(el+' '+state)

            try:    # numpy version >=1.9 required
                nH_vals, cnts = np.unique(tbl[:,0], return_counts=True)
                if np.any( cnts != cnts[0]) or np.any(tbl[:cnts[0],0]!=tbl[0,0]):
                    raise RuntimeError('Table "%s" of unexpected format!' % filename)
                T_vals,  cnts = np.unique(tbl[:,1], return_counts=True)
                if np.any(cnts!=cnts[0]) or cnts[0]!=len(nH_vals) or len(cnts)!=len(T_vals):
                    raise RuntimeError('Table "%s" of unexpected format!' % filename)
            except RuntimeError as rt:
                raise rt
            except:
                nH_vals = np.unique(tbl[:,0])
                T_vals  = np.unique(tbl[:,1])
            tbl = tbl[:,2:].reshape((len(nH_vals),len(T_vals),len(ions)))
        return redshift, nH_vals, T_vals, short_ions, tbl

    def _get_table(self):
        # find all the available tables
        pattern = re.compile( self._pattern.replace('<z>',r'(\d+)') + '$' )
        tables = {}
        for filename in os.listdir(self._tabledir):
            match = re.match(pattern, filename)
            if match:
                z = float(match.group(1)[0]+'.'+match.group(1)[1:])
                tables[z] = os.path.join(self._tabledir,filename)
        # check if any table was found
        if len(tables) == 0:
            raise RuntimeError('No Cloudy tables in the given directory ' +
                               '("%s")!' % self._tabledir)
        # get the two closest tables and interpolate between
        avail_z = np.array( sorted(tables.keys()) )
        i = np.argmin( np.abs( avail_z - self._redshift ) )
        z = avail_z[i]
        if (i==0 and self._redshift<=z) or \
                (i==len(avail_z)-1 and z<=self._redshift):
            # redshift at / over the the limits of the tables
            z1, z2 = avail_z[i], avail_z[i]
        elif z < self._redshift:
            z1, z2 = avail_z[i], avail_z[i+1]
        else:
            z1, z2 = avail_z[i-1], avail_z[i]
        if z1==z2 or z1==self._redshift or z2==self._redshift:
            z = z2 if z2 == self._redshift else z1
            if environment.verbose >= environment.VERBOSE_NORMAL:
                print('load table:')
                print('  "%s" (z=%.3f)' % (tables[z], z))
            redshift, nH_vals, T_vals, ions, table = \
                    self._read_cloudy_table(tables[z])
            if self._style == 'Oppenheimer new':
                assert np.isclose(redshift, z)
                self._nH_vals = nH_vals
                self._T_vals = T_vals
                self._ions = ions
        else:
            if environment.verbose >= environment.VERBOSE_NORMAL:
                print('load tables:')
                print('  "%s" (z=%.3f)' % (tables[z1], z1))
                print('  "%s" (z=%.3f)' % (tables[z2], z2))
            redshift1, nH_vals1, T_vals1, ions1, table1 = \
                    self._read_cloudy_table(tables[z1])
            redshift2, nH_vals2, T_vals2, ions2, table2 = \
                    self._read_cloudy_table(tables[z2])
            if self._style == 'Oppenheimer new':
                z1, z2 = redshift1, redshift2
                if not np.allclose( nH_vals1, nH_vals2 ):
                    raise RuntimeError('nH values in tables do not match!')
                if not np.allclose( T_vals1, T_vals2 ):
                    raise RuntimeError('T values in tables do not match!')
                if not np.all( ions1 == ions2 ):
                    raise RuntimeError('Ions in tables do not match!')
                self._nH_vals = nH_vals1
                self._T_vals  = T_vals1
                self._ions    = ions1
            a = (self._redshift - z1) / (z2 - z1)
            table = (1.-a) * table1 + a * table2

        return table

    def show_table(self, ion=None, cmap='Bright'):
        '''
        Plot the table of the given ion.

        Args:
            ion (str, int):     The ion to plot the table for. If None, all tables
                                are plotted (and a list is returned).
            cmap (str):         The colormap to use.

        Returns:
            fig (Figure):       The figure of the axis plotted on.
            ax (AxesSubplot):   The axis plotted on.
            im (AxesImage):     The image instance created.
            cbar (Colorbar):    The colorbar.
        '''
        import matplotlib as mpl
        from ..plotting import show_image
        from mpl_toolkits.axes_grid1.inset_locator import inset_axes

        if ion is None:
            plts = []
            for ion in self._ions:
                plts.append( self.show_table(ion, cmap=cmap) )
            return plts
        ionname = 'ion'
        if isinstance(ion, str):
            ion = str(ion)
            ionname = ion
            ion = self._ions.index(ion)

        fontsize = 16

        # plot the table as an image
        fig, ax, im = show_image(self._table[:,:,ion],
                                 extent=[[self._nH_vals.min(),self._nH_vals.max()],
                                         [self._T_vals.min(),self._T_vals.max()]],
                                 cmap=cmap)
        ax.set_xlabel(r'$\log_{10}(n\ [%s])$' % Unit('cm**-3').latex(),
                      fontsize=fontsize)
        ax.set_ylabel(r'$\log_{10}(T\ [%s])$' % Unit('K').latex(),
                      fontsize=fontsize)

        # add a colorbar
        clim = [self._table[:,:,ion].min(), self._table[:,:,ion].max()]
        cbar = add_cbar(ax, r'$\log_{10}(f_\mathrm{%s})$' % ionname,
                        clim=clim, cmap=cmap, fontsize=fontsize,
                        nticks=7)


        return fig, ax, im, cbar




# these packages are loaded by gcache3 by default
import os
import io
import sys
import re
import time
import numpy as np
import matplotlib.pyplot as plt
import pygad as pg


# additional packages should be loaded in prepare section
# here your code of functions ONLY starts
def split_cols(col_names):

    col_list = col_names.split(',')
    return col_list

def write_header(fstarform):
    print('%-12s \t %-11s \t %-18s \t %-10s \t %-11s' % (
        'ID', 'a_form', 'r [kpc]', 'r/R200(a)', 'Z'), file=fstarform)

def write_star_form(snapfile, last_snap, last_R200, fstarform):
    '''Store the star formation information.'''
    if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
        print('gather newly formed stars...')
        sys.stdout.flush()

    stars = last_snap.snapshot.stars
    if len(stars) == 0:
        return

    assert stars['form_time'].units == 'a_form'
    new_stars = stars[(snapfile.snapshot.scale_factor < stars['form_time']) &
                      (stars['form_time'] <= last_snap.snapshot.scale_factor)]

    if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
        print('  %d new star particles born' % len(new_stars))

    last_R200_f = float(last_R200)
    this_R200_f = float(snap_cache.gx_properties['R200'])
    last_a = float(last_snap.snapshot.scale_factor)
    this_a = float(snap_cache.snapshot.scale_factor)
    IDs = new_stars['ID'].view(np.ndarray)
    r = new_stars['r'].view(np.ndarray)
    a = new_stars['form_time'].view(np.ndarray)
    x = (a - last_a) / (this_a - last_a)
    R200_a = x * this_R200_f + (1.0-x) * last_R200_f
    del x
    r_in_R200 = r / R200_a
    r_in_kpc = new_stars['r'].in_units_of('kpc',subs=last_snap.snapshot).view(np.ndarray)
    Z = new_stars['metallicity'].view(np.ndarray)
    if pg.environment.verbose >= pg.environment.VERBOSE_TACITURN:
        print('  write their properties to "%s"...' % star_form_filename)
        sys.stdout.flush()
    print("number of new stars ", len(new_stars))
    for i in range(len(new_stars)):
        print('%-12d \t %-11.9f \t %-18.5g \t %-10.5g \t %-11.9f' % (
                    IDs[i], a[i], r_in_kpc[i], r_in_R200[i], Z[i]), file=fstarform)
    return



##############################################################################
# here start the gcache3 sections, invoking your code                        #
##############################################################################
# the order of the gcache3 sections is : init, prepare, process, close, end  #
# sections are identified by line starting with # gcache3-<section name>:    #
# the order must not be modified                                             #
##############################################################################

##############################################################################
# gcache3-init: ignored by gcache3
##############################################################################
pg.environment.verbose=pg.environment.VERBOSE_QUIET
os.environ['SNAPSHOT_HOME'] = '/mnt/hgfs/AstroDaten/CosmoZooms/'
os.environ['SNAPCACHE_HOME'] = '/mnt/hgfs/Astro/CosmoCache/'

snap_fname_rel = "M0094/AGN/snap_m0094_sf_x_2x_094"
snap_cache = pg.SnapshotCache(snap_fname_rel, profile='FastRun')
snap_cache.load_snapshot()

snap_num = 94
snap_start = 94
snap_end = 94
snap_overwrite = True
cmd_par = ""
last_snap = None

##############################################################################
# gcache3-prepare: processes before first snapshot
##############################################################################
print("command CmdStarform - prepare ", snap_num, " von ", snap_start, " bis ", snap_end)

first_snapshot = True

last_R200 = 0.0

##############################################################################
# gcache3-process: processes for each snapshot
##############################################################################

if cmd_par == '':
    fname = "star_form.ascii"
else:
    fname = cmd_par

star_form_filename = snap_cache.get_profile_path() + fname

if snap_overwrite and snap_num ==  snap_start:
    fstarform = open(star_form_filename, "w")
    write_header(fstarform)
else:
    fstarform = open(star_form_filename, "a+")

print("command CmdStarform - write star forming file ", star_form_filename)
if last_snap is None:
    last_R200 = snap_cache.gx_properties['R200']
else:
    write_star_form(snap_cache, last_snap, last_R200, fstarform)
    last_R200 = snap_cache.gx_properties['R200']

snap_cache.gx_properties.append('starformingfile', fname)

first_snapshot = False

##############################################################################
# gcache3-close: processed after last snapshot
##############################################################################
fstarform.close()
print("command CmdStarform - close last_R200=", last_R200)

##############################################################################
# gcache3-end: ignored by gcache3
##############################################################################





'''
A fast octree class (implemented in C, here is only the interface)

TODO:
    * handle UnitArr's! (save units of stored max_H?!)
    * find a way *not* to restrict the numbers of neighbours returned in
      `find_ngbs_within` and `find_ngbs_SPH`
    * implement 2D tree (quadtree)? -> e.g. speed-up for maps

Doctests:
    Generate N random points in a box of [0,L]**3
    Keep N small, since there is a comparison with brute force done!
    >>> N, L, N_ngbs = int(3e4), 1.3, 100
    >>> pos = L * np.random.random((N,3))

    >>> ngbs = N_ngbs * (0.9 + 0.2*np.random.random(N))
    >>> H = L * ( ngbs/N / (4./3*np.pi) )**(1/3.)
    >>> tree = cOctree(pos, H)

    some very basic testing of the interfacing
    >>> x_min, x_max = pos.min(axis=0), pos.max(axis=0)
    >>> center, side_2 = (x_max+x_min)/2.0, max(x_max-x_min)/2.0
    >>> assert np.linalg.norm(tree.center-center)/L<1e-6
    >>> assert abs(tree.side_2-side_2)/L<1e-6
    >>> assert tree.tot_num_part == N
    >>> assert tree.count_nodes(False) <= N
    >>> assert tree.is_in_node(center)
    >>> assert tree.max_H == np.max(H)
    >>> assert not tree.is_in_node(center+[side_2*2.0]*3)
    >>> assert not tree.is_in_node(center-[side_2*2.0]*3)

    >>> assert tree.root is tree
    >>> particles, nodes = 0, 0
    >>> for o in range(8):
    ...     child = tree.get_child(o)
    ...     assert tree.get_octant(child.center) == o
    ...     assert child.parent is tree
    ...     assert child.root is tree
    ...     for o in range(8):
    ...         grandchild = child.get_child(o)
    ...         assert grandchild.parent is child
    ...         assert grandchild.root is tree
    ...     assert tree.is_in_node(child.center)
    ...     particles += child.tot_num_part
    ...     nodes += child.count_nodes()
    >>> assert particles == tree.tot_num_part
    >>> assert nodes + 1 == tree.count_nodes()

    brute force neighbours
    >>> r = center + np.array([-0.3,0.1,0.2]) * side_2
    >>> h = np.mean(H)
    >>> ngbs = []
    >>> for i in range(N):
    ...     d = np.linalg.norm(pos[i] - r)
    ...     if d < h:
    ...         ngbs.append(i)
    >>> assert 0 < len(ngbs) < 3*N_ngbs
    >>> tree_ngbs = tree.find_ngbs_within(r, h, pos, np.inf, max_ngbs=3*N_ngbs)
    >>> assert set(tree_ngbs) == set(ngbs)
    >>> ngbs = []
    >>> for i in range(N):
    ...     d = np.linalg.norm(pos[i] - r)
    ...     if d < H[i]:
    ...         ngbs.append(i)
    >>> assert 0 < len(ngbs) < 3*N_ngbs
    >>> tree_ngbs = tree.find_ngbs_SPH(r, H, pos, np.inf, max_ngbs=3*N_ngbs)
    >>> assert set(tree_ngbs) == set(ngbs)

    Test periodic neighbour finding
    >>> N_p_ngbs = 10
    >>> h = (float(N_p_ngbs)/N_ngbs)**(1/3.) * h
    >>> tree_ngbs = tree.find_ngbs_within([0]*3, h, pos, L, max_ngbs=3*N_ngbs)
    >>> from scipy.spatial.distance import cdist
    >>> normal_dists = cdist(pos[tree_ngbs], [[0,0,0]])
    >>> periodic_dists = np.min( cdist(pos[tree_ngbs], [[0,0,0],
    ...                                                 [0,0,L],
    ...                                                 [0,L,0],
    ...                                                 [0,L,L],
    ...                                                 [L,0,0],
    ...                                                 [L,0,L],
    ...                                                 [L,L,0],
    ...                                                 [L,L,L]]),
    ...                          axis=1)
    >>> assert np.all(periodic_dists<h)
    >>> assert np.any(normal_dists>h)

    More neighbour finding
    >>> pos = np.array([[0.1,0.3,0.2], [0.9,0.3,0.2], [0.8,0.5,0.1],
    ...                 [0.1,0.6,0.8], [0.2,0.2,0.3], [0.6,0.6,0.7]])
    >>> tree = cOctree(pos)
    >>> idx = np.arange(len(pos))

    TODO: the conditioning does not work (properly) in the bitbucket Pipeline
    >>> tree.find_ngbs_within(pos[0], 0.2, pos) #, cond=idx!=0)
    array([0, 4], dtype=uint64)
    >>> tree.find_ngbs_within(pos[0], 0.2, pos, cond=None)
    array([0, 4], dtype=uint64)
    >>> tree.find_next_ngb([0.5]*3, pos)
    5
    >>> tree.find_next_ngb([0.5]*3, pos, cond=idx%2==0)
    4
    >>> tree.find_next_ngb([0.5]*3, pos)#, cond=np.zeros(len(pos)))
    5
'''
__all__ = ['cOctree']

from ..C import *
import sys
import numpy as np
import warnings
import weakref
from .. import environment
from .. import utils

cpygad.new_octree_from_pos.restype = c_void_p
cpygad.new_octree_from_pos.argtypes = [c_size_t, c_void_p]
cpygad.free_octree.argtypes = [c_void_p]
cpygad.get_octree_center.argtypes = [c_void_p, c_void_p]
cpygad.get_octree_side_2.restype = c_double
cpygad.get_octree_side_2.argtypes = [c_void_p]
cpygad.get_octree_is_leaf.restype = c_int
cpygad.get_octree_is_leaf.argtypes = [c_void_p]
cpygad.get_octree_num_children.restype = c_uint
cpygad.get_octree_num_children.argtypes = [c_void_p]
cpygad.get_octree_tot_part.restype = c_size_t
cpygad.get_octree_tot_part.argtypes = [c_void_p]
cpygad.get_octree_max_H.restype = c_double
cpygad.get_octree_max_H.argtypes = [c_void_p]
cpygad.get_octree_max_H.restype = c_double
cpygad.get_octree_max_H.argtypes = [c_void_p]
cpygad.get_octree_max_depth.restype = c_int
cpygad.get_octree_max_depth.argtypes = [c_void_p]
cpygad.get_octree_node_count.restype = c_size_t
cpygad.get_octree_node_count.argtypes = [c_void_p, c_int]
cpygad.get_octree_in_region.restypes = c_int
cpygad.get_octree_in_region.argtypes = [c_void_p, c_void_p]
cpygad.fill_octree.argtypes = [c_void_p, c_size_t, c_void_p]
cpygad.update_octree_max_H.argtypes = [c_void_p, c_void_p]
cpygad.update_octree_const_max_H.argtypes = [c_void_p, c_double]
cpygad.get_octree_child.restype = c_void_p
cpygad.get_octree_child.argtypes = [c_void_p, c_int]
cpygad.get_octree_octant.restype = c_uint
cpygad.get_octree_octant.argtypes = [c_void_p, c_void_p]
cpygad.get_octree_ngbs_within.argtypes = [c_void_p,
                                          c_void_p, c_double,
                                          c_size_t, c_void_p, POINTER(c_size_t),
                                          c_void_p, c_double,
                                          c_void_p]
cpygad.get_octree_ngbs_SPH.argtypes = [c_void_p,
                                       c_void_p, c_void_p,
                                       c_size_t, c_void_p, POINTER(c_size_t),
                                       c_void_p, c_double, c_double]
cpygad.get_octree_next_ngb.argtypes = [c_void_p, c_void_p, c_void_p, c_double,
                                       c_void_p]


class _MAX_TREE_LEVEL_class(type):
    _MAX_TREE_LEVEL = int(c_int.in_dll(cpygad, 'MAX_TREE_LEVEL').value)

    def _get_MAX_TREE_LEVEL(self):
        return self._MAX_TREE_LEVEL

    def _set_MAX_TREE_LEVEL(self, value):
        raise AttributeError("can't set attribute")

    MAX_TREE_LEVEL = property(_get_MAX_TREE_LEVEL, _set_MAX_TREE_LEVEL)


class cOctree(object, metaclass=_MAX_TREE_LEVEL_class):
    '''
    A octree implementation with the backend in written in C.

    Actually this is a wrapper to the C++ template class Tree<3>. Internally only
    indices are stored so that any property of a particle can be referenced.

    Args:
        center (array-like):    TODO
        side_2 (float):         TODO
    '''

    @property
    def MAX_TREE_LEVEL(self):
        return cOctree.MAX_TREE_LEVEL

    def __init__(self, pos, H=None):
        if environment.verbose >= environment.VERBOSE_TALKY:
            print('build a cOctree with %s positions' % (
                utils.nice_big_num_str(len(s))))
            sys.stdout.flush()
        pos = np.asarray(pos, dtype=np.float64)
        if pos.shape[1:] != (3,):
            raise ValueError('Positions have to have shape (N,3)!')
        if pos.base is not None:
            pos = pos.copy()

        self.__parent = None
        self.__node_ptr = cpygad.new_octree_from_pos(len(pos), pos.ctypes.data)
        # If this object does not have any references anymore, and hence will get
        # garbage collected, the weakref will not reference any object anymore
        # and, hence, call its callback function, which in turn has the
        # responsibility to free the underlying C memory. We need to make sure the
        # closure does not hold references to `self`. Otherwise the weakref would
        # always point to a living object, and thus would never call its callback
        # function.
        # Note: child node cOctrees do not get instantiated with a call of
        # `__init__`.
        self.__weakref_for_freeing_C_memory = weakref.ref(
            self,
            lambda wr, ptr=self.__node_ptr: cpygad.free_octree(ptr),
        )

        if H is not None:
            self.update_max_H(H)

        if environment.verbose >= environment.VERBOSE_TALKY:
            print('done.')
            sys.stdout.flush()

    @property
    def parent(self):
        '''The parent node of this node (None for root).'''
        return self.__parent

    @property
    def root(self):
        '''The root of this octree node (self for root node).'''
        root = self
        while root.__parent:
            root = root.__parent
        return root

    @property
    def center(self):
        center = np.empty((3,), dtype=np.float64)
        cpygad.get_octree_center(self.__node_ptr, center.ctypes.data)
        return center

    @property
    def side_2(self):
        '''Half the side length of the tree box.'''
        return float(cpygad.get_octree_side_2(self.__node_ptr))

    @property
    def full_side(self):
        '''The total side length of the tree box.'''
        return 2.0 * self.side_2

    @property
    def is_leaf(self):
        '''Whether this is a leaf node.'''
        return bool(cpygad.get_octree_is_leaf(self.__node_ptr))

    @property
    def num_children(self):
        '''Number of child nodes.'''
        return int(cpygad.get_octree_num_children(self.__node_ptr))

    @property
    def tot_num_part(self):
        '''Get the total number of particles in the tree.'''
        return int(cpygad.get_octree_tot_part(self.__node_ptr))

    @property
    def max_H(self):
        '''The maximum smoothing length (as support radius) in the tree.'''
        return float(cpygad.get_octree_max_H(self.__node_ptr))

    @property
    def max_depth(self):
        '''The maximum depth of this particular tree. (Only node would be 0.)'''
        return int(cpygad.get_octree_max_depth(self.__node_ptr))

    def get_child(self, o):
        '''Return the child of octant o.'''
        if 0 <= o < self.num_children:
            # Do not call __init__ and hence do not set the weakref, which would
            # free the memory of this child, which is part of the parent's tree!
            child = cOctree.__new__(cOctree)
            child.__parent = self
            child.__node_ptr = cpygad.get_octree_child(self.__node_ptr, o)
            return child
        else:
            return None

    def get_octant(self, r):
        '''Get the octant of position `r`.'''
        r = np.asarray(r, dtype=np.float64).copy()
        if r.shape != (3,):
            raise ValueError('Position has to have shape (3,)!')
        return int(cpygad.get_octree_octant(self.__node_ptr, r.ctypes.data))

    def count_nodes(self, count_non_leaves=True):
        '''
        Count all nodes in the tree.

        Args:
            count_non_leaves (bool):    Whether to count all nodes (that is also
                                        include nodes that are no leaves). If set
                                        to False, only count leaf nodes.

        Returns:
            nodes (int):                The number of nodes.
        '''
        nol = int(bool(count_non_leaves))
        return int(cpygad.get_octree_node_count(self.__node_ptr, nol))

    def is_in_node(self, r):
        '''Check whether position r lies within this node.'''
        r = np.asarray(r, dtype=np.float64).copy()
        if r.shape != (3,):
            raise ValueError('Position has to have shape (3,)!')
        return bool(cpygad.get_octree_in_region(self.__node_ptr, r.ctypes.data))

    def update_max_H(self, H=None):
        '''
        Fill the tree with particles (without deleting former particles from
        tree).

        Args:
            H (array-like, float):  The smoothing lengthes of the particles. Has
                                    to have shape (N,) or can be a float (all
                                    smoothing lengthes then are the same).
        '''
        from numbers import Number
        if isinstance(H, Number):
            H = float(H)
            cpygad.update_octree_const_max_H(self.__node_ptr, H)
        else:
            H = np.asarray(H, dtype=np.float64)
            if H.shape != (self.tot_num_part,):
                raise ValueError('Smoothing lengthes have to have shape (N,)!')
            if H.base is not None:
                H = H.copy()
            cpygad.update_octree_max_H(self.__node_ptr, H.ctypes.data)

    def find_ngbs_within(self, r, H, pos, periodic=np.inf, cond=None, max_ngbs=100):
        '''
        Find all particles in tree within distance `H` from position `r`.

        Args:
            r (array-like):     Reference position.
            H (float):          Maximum distance to search for neighbours in.
            pos (array-like):   The positions corresponding to the indices of the
                                tree.
            periodic (float):   Assume the particles to sit in a periodic cube
                                with this side length.
            cond (array-like):  Boolean array of a condition to be fulfilled such
                                that a particle is registered as a neighbour. In
                                other words, each particle i for which cond[i] is
                                False is excluded from the neighbours.
            max_ngbs (int):     Only return this number of neighbours at maximum.
                                (Not necessarily the closest ones!)

        Returns:
            ngbs (np.ndarray):  List if the indices of the neighbours (in random
                                order). At maximum, though, `max_ngbs` of them.
        '''
        r = np.asarray(r, dtype=np.float64).copy()
        H = float(H)
        pos = np.asarray(pos, dtype=np.float64)
        if pos.shape[1:] != (3,):
            raise ValueError('Positions have to have shape (N,3)!')
        if pos.base is not None:
            pos = pos.copy()
        max_ngbs = int(max_ngbs)
        periodic = float(periodic)

        ngbs = np.empty(max_ngbs, dtype=np.uintp)
        N_ngbs = c_size_t()
        if cond is not None:
            cond = np.asarray(cond, dtype=np.int32)
            if cond.shape != (len(pos),):
                raise ValueError('Unmatching shape of `cond`: %s!' % (cond.shape,))
            if cond.base is not None:
                cond = cond.copy()
            cond = cond.ctypes.data
        cpygad.get_octree_ngbs_within(self.__node_ptr,
                                      r.ctypes.data, H,
                                      max_ngbs, ngbs.ctypes.data, byref(N_ngbs),
                                      pos.ctypes.data, periodic,
                                      cond,
                                      )
        #ngbs.resize(N_ngbs.value)
        erg = np.resize(ngbs, N_ngbs.value)
        return erg

    def find_ngbs_SPH(self, r, H, pos, periodic=np.inf, max_ngbs=100):
        '''
        Find all particles in tree within distance `H` from position `r`.

        Args:
            r (array-like):     Reference position.
            H (array-like):     The smoothing lengthes (support radius)
                                corresponding to the indices of the tree.
            pos (array-like):   The positions corresponding to the indices of the
                                tree.
            periodic (float):   Assume the particles to sit in a periodic cube
                                with this side length.
            max_ngbs (int):     Only return this number of neighbours at maximum.
                                (Not necessarily the closest ones!)

        Returns:
            ngbs (np.ndarray):  List if the indices of the neighbours (in random
                                order). At maximum, though, `max_ngbs` of them.
        '''
        r = np.asarray(r, dtype=np.float64).copy()
        pos = np.asarray(pos, dtype=np.float64)
        if pos.shape[1:] != (3,):
            raise ValueError('Positions have to have shape (N,3)!')
        H = np.asarray(H, dtype=np.float64)
        if H.shape != (len(pos),):
            raise ValueError('Positions have to have shape (N,)!')
        if pos.base is not None:
            pos = pos.copy()
        if H.base is not None:
            H = H.copy()
        max_ngbs = int(max_ngbs)
        periodic = float(periodic)

        ngbs = np.empty(max_ngbs, dtype=np.uintp)
        N_ngbs = c_size_t()
        cpygad.get_octree_ngbs_SPH(self.__node_ptr,
                                   r.ctypes.data, H.ctypes.data,
                                   max_ngbs, ngbs.ctypes.data, byref(N_ngbs),
                                   pos.ctypes.data,
                                   periodic,
                                   0.0,
                                   )
        #ngbs.resize(N_ngbs.value)
        erg = np.resize(ngbs, N_ngbs.value)
        return erg

    def find_next_ngb(self, r, pos, periodic=np.inf, cond=None):
        '''
        Find all particles in tree within distance `H` from position `r`.

        Args:
            r (array-like):     Reference position.
            pos (array-like):   The positions corresponding to the indices of the
                                tree.
            periodic (float):   Assume the particles to sit in a periodic cube
                                with this side length.
            cond (array-like):  Boolean array of a condition to be fulfilled such
                                that a particle is registered as a neighbour. In
                                other words, each particle i for which cond[i] is
                                False is excluded from the neighbours.

        Returns:
            ngb (int):          Index of the next neighbour to position r that
                                fulfills the condition. None if there is no such
                                next neighbour.
        '''
        r = np.asarray(r, dtype=np.float64).copy()
        pos = np.asarray(pos, dtype=np.float64)
        if pos.shape[1:] != (3,):
            raise ValueError('Positions have to have shape (N,3)!')
        if pos.base is not None:
            pos = pos.copy()
        periodic = float(periodic)

        if cond is not None:
            cond = np.asarray(cond, dtype=np.int32)
            if cond.shape != (len(pos),):
                raise ValueError('Unmatching shape of `cond`: %s!' % (cond.shape,))
            if cond.base is not None:
                cond = cond.copy()
            cond = cond.ctypes.data
        ngb = cpygad.get_octree_next_ngb(self.__node_ptr,
                                         r.ctypes.data,
                                         pos.ctypes.data,
                                         periodic,
                                         cond,
                                         )
        if ngb == -1:
            ngb = None

        return ngb




"""
Module for units and arrays with units.

Also doctest other parts of this sub-module:
    >>> import doctest
    >>> doctest.testmod(units)
    TestResults(failed=0, attempted=38)
    >>> doctest.testmod(unit_arr)
    TestResults(failed=0, attempted=85)
"""
import os

from ..environment import module_dir
from .unit_arr import *
from .units import *

define_from_cfg(
    [
        "./units.cfg",
        os.getenv("HOME") + "/.config/pygad/units.cfg",
        module_dir + "config/units.cfg",
    ]
)



'''
A octree implementation that stores indices rather than positions itself. This
approach saves memory and makes it possible to also refere to other data attached
to the points as long as it is appropriately stored in an array, too (like in
snapshots).

Trees, subtrees, and nodes are all represented by the same OctNode class.

Note:
    This is a pure Python implementation and, hence, not very fast! However, an
    octree gives some more spatial information than the bare cloud of points.

Example:
    Start with a regular grid of points:
    >>> pos = np.linspace(0,1, 32)
    >>> pos = np.vstack(map(np.ravel, np.meshgrid(pos, pos, pos))).T
    >>> OctNode.SPLIT_NUM = 12
    >>> tree = Octree(pos)
    >>> count_nodes(tree)
    4681
    >>> count_leaves(tree)
    4096
    >>> max_depth(tree)
    4
    >>> tree.get_child([7,0,5,3])
    OctNode(center=[0.281,0.4062,0.3436], side=0.06256, depth=4, leaf)
    >>> tree.get_child([7,0,5,3]).child
    [12554, 12555, 12586, 12587, 13578, 13579, 13610, 13611]
    >>> find_node(tree, pos[12586]) is tree.get_child([7,0,5,3])
    True
    >>> for r in pos:
    ...     n = find_node(tree, r)
    ...     if max(abs(r - n.center)) > n.side:
    ...         print(r, n, n.side)
    >>> for i in range(20):
    ...     r = np.random.rand(3)
    ...     np_idcs = np.where(cdist(pos, [[.3,.3,.3]]).ravel() < 0.1)[0]
    ...     ot_idcs = find_ngbs(tree, [.3,.3,.3], 0.1, pos)
    ...     if not set(np_idcs) == set(ot_idcs):
    ...         print('numpy:')
    ...         print(pos[np_idcs])
    ...         print( 'octree:')
    ...         print(pos[ot_idcs])

    A more realistic sample of positions (just every 100th for speed reasons):
    >>> from ..snapshot.snapshot import Snapshot
    >>> from ..environment import module_dir
    >>> snap = Snapshot(module_dir+'snaps/snap_M1196_4x_470', physical=False,
    ...             load_double_prec=True)
    >>> snap = snap[::100]
    >>> OctNode.MAX_DEPTH = 11  # making the following tree build way faster
    >>> tree = Octree(snap)
    load block pos... done.
    >>> count_particles(tree), count_nodes(tree), count_leaves(tree)
    (20791, 5515, 4736)
    >>> apply_on(tree, delete_where_too_many_heavy_particles(snap,
    ...                                     min_side='50 ckpc / h_0'))
    (0, 18643)
    >>> sub = snap[all_indices(tree)]
    >>> sub.parts
    [8573, 9334, 0, 0, 736, 0]
    >>> apply_on(tree, populate_center_of_mass(snap['pos'], snap['mass']))  # doctest:+ELLIPSIS
    load block mass... done.
    (array([34497.34533395, 35431.564173  , 33097.62806547]), 2.926226872834377)
    >>> from .. import analysis
    >>> com = analysis.center_of_mass(sub)
    >>> d = np.sqrt(np.sum((tree.com - com)**2))
    >>> if not d < 0.5:
    ...     print('com      =', com)
    ...     print('tree.com =', tree.com)
    >>> del_quantity(tree, 'com')
    >>> del_quantity(tree, 'mass')
'''
__all__ = ['OctNode', 'Octree', 'count_nodes', 'count_leaves', 'count_particles',
           'max_depth', 'find_node', 'find_ngbs', 'all_indices', 'apply_on',
           'del_quantity', 'populate_center_of_mass',
           'delete_where_too_many_heavy_particles']

import numpy as np
from scipy.spatial.distance import cdist
from ..units import *
from .. import environment
import sys


def find_octant(pos, ref):
    '''Index corresponding to an octant of 'pos' with respect to 'ref'.'''
    res = 0
    if pos[0] < ref[0]: res += 1
    if pos[1] < ref[1]: res += 2
    if pos[2] < ref[2]: res += 4
    return res


class OctNode(object):
    '''
    Class for an octree node.

    You should not create it directly, but via the factory function Octree. The
    data stored per leaf is not the positions of the individual particle, but
    rather a reference to an array holding all positions (and possible more)
    together with the indices of the particles of that leaf. This saves memory and
    enables one to also refere to other data of the particle stored somewhere
    else via the indices.

    Note:
        This class does not use data encapsulation for convenience.
        Be responsible!

    Static variables:
        MAX_DEPTH (int):        The maximum depth to use for the octree. It is
                                neccessary to avoid stack overflows in cases where
                                two particles are very close to each other.
        SPLIT_NUM (int):  The number of particles at which to split a leaf
                                node. Hence, the maximum number of particles per
                                (leaf) node is SPLIT_NUM-1.
    '''

    MAX_DEPTH = 15
    SPLIT_NUM = 7

    def __init__(self, center, side, depth):
        self.center = center  # the center of this octree node cube
        self.side = side  # the (full) side length of this cube
        self.depth = depth  # the depth level of this node (root is 0)
        self.is_leaf = True  # whether this node has no children
        self.child = []  # either the child nodes or the indices of the
        # positions if this is a leaf node

    def __repr__(self):
        re = 'OctNode('
        re += 'center=[%.4g,%.4g,%.4g]' % tuple(coord for coord in self.center)
        re += ', side=%.4g' % self.side
        re += ', depth=%d' % self.depth
        if self.is_leaf:
            re += ', leaf'
        else:
            re += ', non-leaf'
        re += ')'
        return re

    def particles(self):
        '''Count the number of particles in the (sub-)tree of this node.'''
        if self.is_leaf:
            return len(self.child)
        cnt = 0
        for child in self.child:
            if child is not None:
                cnt += child.particles()
        return cnt

    def get_child(self, path):
        '''
        Get a child by following the path of octants.

        Args:
            path (list):    The child octant number path to follow. Example:
                            path=[3,0,7] would give the child of octant 7 of the
                            the child of octant 0 of the child of octant 3 of this
                            node.

        Returns:
            child (OctNode, ...):   The corresponding child.
        '''
        if len(path) == 1:
            return self.child[path[0]]
        else:
            return self.child[path[0]].get_child(path[1:])

    def insert(self, idx, pos_data):
        '''
        Insert a data point into this node.

        Note:
            Using a UnitArr as pos_data can be slow, since it is indexed in only
            the first dimension, leaving the units. consider passing a np.ndarray
            view.

        Args:
            idx (int):              The index of the position in the position
                                    vector pos_data.
            pos_data (array-like):  All positions. It has to have shape (N,3) ---
                                    an array of 3-dim. position vectors. However,
                                    it is not neccessary that all positions of
                                    this array are inserted to the tree.
        '''
        if self.is_leaf:
            self.child.append(idx)
            if len(self.child) > OctNode.SPLIT_NUM \
                    and self.depth < OctNode.MAX_DEPTH:
                self.is_leaf = False
                idcs = self.child
                self.child = [None] * 8
                for idx in idcs:
                    self.insert(idx, pos_data)
        else:
            octant = find_octant(pos_data[idx], self.center)
            if self.child[octant] is None:
                cntr = self.center
                off = self.side / 4.0
                if octant == 0:
                    cntr = (cntr[0] + off, cntr[1] + off, cntr[2] + off)
                elif octant == 1:
                    cntr = (cntr[0] - off, cntr[1] + off, cntr[2] + off)
                elif octant == 2:
                    cntr = (cntr[0] + off, cntr[1] - off, cntr[2] + off)
                elif octant == 3:
                    cntr = (cntr[0] - off, cntr[1] - off, cntr[2] + off)
                elif octant == 4:
                    cntr = (cntr[0] + off, cntr[1] + off, cntr[2] - off)
                elif octant == 5:
                    cntr = (cntr[0] - off, cntr[1] + off, cntr[2] - off)
                elif octant == 6:
                    cntr = (cntr[0] + off, cntr[1] - off, cntr[2] - off)
                elif octant == 7:
                    cntr = (cntr[0] - off, cntr[1] - off, cntr[2] - off)
                cntr = np.array(cntr)
                self.child[octant] = OctNode(cntr, self.side / 2.0, self.depth + 1)
            self.child[octant].insert(idx, pos_data)


def Octree(pos_data, idx=None, center=None, world_size=None):
    '''
    Build an octree from position data.

    Args:
        pos_data (UnitArr, np.ndarray):     A (N,3)-array of positions.
        idx (array-like):                   The indices of the particles/positions
                                            in pos_data to add to the tree. If
                                            None, all particles are added.
        center (UnitArr, array-like):       The center of the octree. If omitted,
                                            the middle of all position is taken.
        world_size (UnitArr, Unit, float):  The size (full side length) of the
                                            root node of the octree. If not
                                            specified, the maximum distance in any
                                            coordinate of pos_data from the center
                                            is taken.

    Returns:
        root (OctNode):     The root node of the octree built.
    '''
    from ..snapshot.snapshot import Snapshot
    if isinstance(pos_data, Snapshot):
        s = pos_data
        bz = float(s.boxsize.in_units_of(s['pos'].units, copy=False))
        return Octree(s['pos'], center=[bz / 2.] * 3, world_size=bz)

    if environment.verbose >= environment.VERBOSE_TALKY:
        print('build a Octree with %s positions' % (
            utils.nice_big_num_str(len(s))))
        sys.stdout.flush()

    if center is None:
        upper = np.max(pos_data, axis=0)
        lower = np.min(pos_data, axis=0)
        center = (upper + lower) / 2.0

    if world_size is None:
        try:
            world_size = 1.001 * pos_data.ptp(axis=0).max()
        except:
            world_size = 1.001 * np.array(pos_data).ptp(axis=0).max()

    if isinstance(pos_data, UnitArr):
        units = pos_data.units
        if hasattr(center, 'in_units_of'):
            center = center.in_units_of(units, copy=False)
        if hasattr(world_size, 'in_units_of'):
            world_size = world_size.in_units_of(units, copy=False)
        pos_data = pos_data.view(np.ndarray)
    pos_data = np.array(pos_data)

    root = OctNode(np.array(list(center)), float(world_size), 0)

    if idx is None:
        idx = range(len(pos_data))
    for i in idx:
        root.insert(i, pos_data)

    if environment.verbose >= environment.VERBOSE_TALKY:
        print('done.')
        sys.stdout.flush()

    return root


def count_nodes(node):
    '''Count the total number of nodes within the given tree/node.'''
    cnt = 1  # there is at least this node
    if not node.is_leaf:
        for child in node.child:
            if child is not None:
                cnt += count_nodes(child)
    return cnt


def count_leaves(node):
    '''Count the number of leaf nodes within the given tree/node.'''
    if node.is_leaf:
        return 1
    cnt = 0
    for child in node.child:
        if child is not None:
            cnt += count_leaves(child)
    return cnt


def count_particles(node):
    '''Count the number of particles within the given tree/node.'''
    if node.is_leaf:
        return len(node.child)
    cnt = 0
    for child in node.child:
        if child is not None:
            cnt += count_particles(child)
    return cnt


def max_depth(node):
    '''Find the maximum depth within the given tree/node.'''
    if node.is_leaf:
        return node.depth
    d_max = -np.inf
    for child in node.child:
        if child is not None:
            d_max = max(d_max, max_depth(child))
    return d_max


def find_node(tree, pos):
    '''
    Find the 'smallest' node that would contain the position.

    Args:
        tree (OctNode):         The root node of the tree to search in.
        pos (array-like):       The reference position to find the node for.

    Returns:
        node (OctNode):         The smallest node (i.e. the one with the highest
                                depth) that contains the position. It is not
                                neccessarily a leaf node; it might be a node that
                                has children, but not the one that would contain
                                the position of interest.
    '''
    if tree.is_leaf:
        return tree
    o = find_octant(pos, tree.center)
    if tree.child[o] is None:
        return tree
    return find_node(tree.child[o], pos)


def _find_ngbs(node, pos, r, pos_data):
    '''Helper function for speed. See find_ngbs!'''
    if node.is_leaf:
        d = cdist(pos_data[node.child], [pos]).ravel()
        return list(np.array(node.child)[d <= r])
    # half the size of the cube of a child node (plus some safety margin)
    child_half_side = 1.01 * node.side / 4.0
    ngbs = []
    for child in node.child:
        if child is None:
            continue
        # if the distance to the edge of the child node is too large, also skip
        if all(abs(child.center - pos) > r + child_half_side):
            continue
        ngbs += _find_ngbs(child, pos, r, pos_data)
    return ngbs


def find_ngbs(node, pos, r, pos_data):
    '''
    Find the all neighbours of the given position that are closer than r.

    Note:
        Since numpy uses C in the backend (e.g. for loops), it is still orders of
        magnidutes faster to just use scipy.spatial.distance.cdist together with
        np.where rather than this function (even if the tree is already built)!

    Args:
        node (OctNode):             The root node of the tree to search in.
        pos (UnitArr, array-like):  The reference position to find neighbours for.
        r (UnitArr, Unit, float):   The maximum distance to the neighbours.
        pos_data (UnitArr, array-like):
                                    The position data of the octree.
    '''
    if isinstance(pos_data, UnitArr):
        units = pos_data.units
        if hasattr(pos, 'in_units_of'):
            pos = pos.in_units_of(units, copy=False)
        if hasattr(r, 'in_units_of'):
            r = float(r.in_units_of(units, copy=False))
        pos_data = pos_data.view(np.ndarray)
    return _find_ngbs(node, pos, r, pos_data)


def _all_indices(node):
    '''Helper function for speed (no copy). See all_indices!'''
    if node.is_leaf:
        return node.child
    else:
        idx = []
        for child in node.child:
            if child is not None:
                idx += _all_indices(child)
        return idx


def all_indices(node):
    '''Gather all position indices within the tree.'''
    if node.is_leaf:
        return np.array(node.child, int)
    else:
        return np.array(_all_indices(node), int)


def apply_on(node, action):
    '''
    Apply the defined action on each node recursively from bottom.

    Args:
        node (OctNode):     The tree on which nodes the action is applied on.
        action (callable):  The action that is applied with the following call
                            signature: action(node, children), where children is a
                            list of the results of all the children (if it does
                            not exist, it is None).
    '''
    children = None
    if not node.is_leaf:
        children = [None] * 8
        for o, child in enumerate(node.child):
            if child is not None:
                children[o] = apply_on(child, action)
    return action(node, children)


def del_quantity(node, name):
    '''Delete a quantity with the given name at each node.'''
    if not node.is_leaf:
        for child in node.child:
            if child is not None:
                del_quantity(child, name)
    delattr(node, name)


class populate_center_of_mass(object):
    '''
    A callable that can be passed to 'apply_on' in order to populate the tree with
    center of mass (attribute 'com') and total mass (attribute 'mass')
    information.

    Args:
        pos (np.array):     The position data of the tree.
        mass (np.array):    The mass data of the tree.

    Call signature:
        fill_center_of_mass(node,children) -> (com,mass)

        where children is an array of the results of the children of this node.
        For children which do not exist the entry is None.
    '''

    def __init__(self, pos, mass):
        # a view as a np.ndarray makes the slicing (indexing in the first
        # dimension) faster
        self.pos = pos.view(np.ndarray)
        self.mass = mass.view(np.ndarray)

    def __call__(self, node, children):
        if node.is_leaf:
            r = self.pos[node.child]
            m = self.mass[node.child]
        else:
            r, m = np.zeros((8, 3)), np.zeros(8)
            for o, child in enumerate(children):
                if child is not None:
                    r[o] = child[0]
                    m[o] = child[1]
        setattr(node, 'mass', m.sum())
        setattr(node, 'com', (m * r.T).sum(axis=1) / node.mass)
        return node.com, node.mass


class delete_where_too_many_heavy_particles(object):
    '''
    A callable that can be passed to 'apply_on' in order to delete all tree leaves
    where the fraction of low resolution-particles is too high.

    Args:
        snap (Snap):        The snapshot with the position data of the tree.
        min_side (UnitScalar):
                            The minimum side length of a node to be allowed to be
                            deleted. If this would be 0.0, propably only leaf
                            nodes containing low resolution particles would be
                            deleted, since OctNode.SPLIT_NUM is typically of order
                            10 and, hence, a single particle brings the fraction
                            up to 10%.
        max_frac (float):   The maximum fraction of los resolution-particles
                            allowed.
        lowres_pts (list):  The particle types to consider low resolution
                            particles.
                            Default: [2,3]

    Call signature:
        delete_where_too_many_heavy_particles(node,children) -> (N_low_res,N_tot)

        where children is an array of the results of the children of this node.
        For children which do not exist the entry is None.
    '''

    def __init__(self, snap, min_side, max_frac=1e-3, lowres_pts=None):
        self.max_frac = max_frac
        self.min_side = float(UnitScalar(min_side, snap.boxsize.units, subs=snap))

        self.lowres_pts = [2, 3] if lowres_pts is None else list(lowres_pts)
        self.N_cumsum = np.cumsum(snap.parts)

        def is_low_res(i):
            for pt in self.lowres_pts:
                if self.N_cumsum[pt - 1] <= i < self.N_cumsum[pt]:
                    return True
            return False

        self.is_low_res_idx = is_low_res

    def __call__(self, node, children):
        if node.is_leaf:
            N_low_res = sum(self.is_low_res_idx(i) for i in node.child)
            N_tot = len(node.child)
            return N_low_res, N_tot
        else:
            N_low_res, N_tot = 0, 0
            for o, child in enumerate(children):
                if child is not None:
                    if node.child[o].side > self.min_side and (child[1] == 0 \
                                                               or float(child[0]) / child[1] > self.max_frac):
                        node.child[o] = None
                    else:
                        N_low_res += child[0]
                        N_tot += child[1]
            return N_low_res, N_tot




'''
Utility functions regarding geometry.
'''
__all__ = ['angle', 'dist', 'find_maxima_prominence_isolation']

import numpy as np
from ..units import UnitArr, UnitQty, UnitQty

def angle(a,b):
    '''
    Return the angle (in radiants) between two 3D vectors.

    Examples:
        >>> angle( [1,2,3], [1,1.23,-1.4] )
        UnitArr(1.664446397854663, units="rad")
        >>> angle( [0,1,0], [1,0,0] ).in_units_of('degree')
        UnitArr(90.0, units="degree")
        >>> angle( [1,0,0], [1,0,0] )
        UnitArr(0.000000e+00, units="rad")
        >>> a = UnitArr([1.23,-1.2,0.3], 'cm')
        >>> b = UnitArr([-0.79,0.43,-2.48], 'cm')
        >>> angle( a, np.cross(a,b) ).in_units_of('degree')
        UnitArr(90.0, units="degree")
    '''
    cos = np.dot(a,b) / ( np.linalg.norm(a) * np.linalg.norm(b) )
    return UnitArr(np.arccos(cos),'rad')

def dist(arr, pos=None, metric='euclidean', p=2, V=None, VI=None, w=None):
    '''
    Calculate the distances of the positions in arr to pos.

    This function uses scipy.spatial.distance.cdist and is, hence, faster than
    sqrt(sum((arr-pos)**2,axis=1)). Also see its documentation for more
    information. This is only a wrapper that handles the units. The overhead is
    marginal.

    Args:
        arr (array-like):   The array of positions (shape: (...,N)).
        pos (array-like):   The reference position (shape: (N,)).
                            Default: [0]*N
        [...]:              See scipy.spatial.distance.cdist.

    Returns:
        dists (UnitArr):    Basically cdist(arr, [pos], [...]).ravel() with units.

    Examples:
        >>> dist( [1,2,0], [-2,0,3] )
        UnitArr([4.69041576])
        >>> dist( UnitArr([1,2,0],'m'), UnitArr([-231,12,323],'cm') )
        UnitArr([4.99233412], units="m")
        >>> dist( UnitArr([1,2,0],'kpc') )
        UnitArr([2.23606798], units="kpc")
        >>> dist( [[1,2,0],[1,2,3],[-2,3,4]] )
        UnitArr([2.23606798, 3.74165739, 5.38516481])
    '''
    from scipy.spatial.distance import cdist
    arr = UnitQty(arr)
    if pos is None:
        if not isinstance(arr, np.ndarray): # includes UnitArr!
            arr = np.array(arr)
        pos = [0]*arr.shape[-1]
    pos = UnitQty(pos)

    if arr.units is not None:
        units = arr.units
        if pos.units is not None:
            pos = pos.astype(float).in_units_of(arr.units)
    elif pos.units is not None:
        units = pos.units
    else:
        units = None

    if arr.ndim == 1:
        arr = arr.reshape((1,-1))

    res = cdist(arr, [pos]).ravel().view(UnitArr)
    res.units = units
    return res

def find_maxima_prominence_isolation(arr, prominence=None, sortby='index',
                                     descending=False):
    '''
    Find all local maxima in an array (of a minimum prominence).

    This function does not only return the maxima, but also their position (as
    an index), their prominence, and their isolation (in pixels/indices). The
    latter two have the usual topographic definitions.

    Note:
        If two or more consequtive values are the same and together build a
        local extremum, they will all count as individual local extrema.

    Args:
        arr (array-like):       The 1-dimensional array in which to look for
                                the maxima.
        prominence (float):     A minimum prominence to filter for.
                                If None, do not filter the results.
        sortby (str):           Whether / how to sort the results. This must be
                                one of 'index', 'value', 'prominence', and
                                'isolation'.
                                They are naturally sorted by the indices. Hence,
                                this option is the fastest when chosen with
                                `descending=False`.
        descending (bool):      Sort in descending order.

    Returns:
        maxima (np.ndarray):    All the (filtered) maxima. It is an array with
                                named fields: 'index', 'value', 'prominence',
                                and 'isolation'.
    Examples:
        >>> m = find_maxima_prominence_isolation( np.array([1,2,3,2,0,2,1,3],dtype=float) )
        >>> m['index']
        array([2, 5, 7])
        >>> m['value']
        array([3., 2., 3.])
        >>> m['prominence']
        array([3., 1., 3.])
        >>> m['isolation']
        array([6, 2, 8])
        >>> find_maxima_prominence_isolation( [1,1,1] )['value']
        array([1, 1, 1])
        >>> find_maxima_prominence_isolation( [0,1,4,1,0] )['index']
        array([2])
        >>> find_maxima_prominence_isolation( [0,1,4,1,-3,-2,1,-1,0,3,5,6,4,1],
        ...                                   sortby='prominence',
        ...                                   descending=True )['value']
        array([6, 4, 1])
        >>> find_maxima_prominence_isolation( [0,5,-2,-2,1,0,8,0] )[0]
        (1, 5, 7, 5)
        >>> len(find_maxima_prominence_isolation( [0] ))
        0
    '''
    arr = np.asarray(arr)
    if arr.ndim != 1:
        raise ValueError('Only one dimensional array can be processed!')
    if len(arr) < 2:
        return np.array([],
                        dtype=[('index',int), ('value',arr.dtype), ('prominence',arr.dtype)])

    neg_arr = -arr
    def is_local_max(i, a=arr):
        next_other, j = a[i], i
        while next_other == a[i]:
            j += 1
            next_other = a[i]-1 if j==len(a) else a[j]
        prev_other, j = a[i], i
        while prev_other == a[i]:
            j -= 1
            prev_other = a[i]-1 if j==-1 else a[j]
        return prev_other < a[i] > next_other
    def is_local_min(i):
        return is_local_max(i, a=neg_arr)

    # find all the local extrema in the smoothed spectrum
    extrema = [ ]
    for i in range(len(arr)):
        if is_local_max(i):
            extrema.append( ['max', arr[i], i] )
        elif is_local_min(i):
            extrema.append( ['min', arr[i], i] )

    # get the prominence of the maxima
    for iex,ex in enumerate(extrema):
        if ex[0] == 'min':
            ex.append( -1.0 )
            continue

        # find the prominence
        def find_one_sided_prominence(iex, ex, left):
            prom_one = np.inf
            vmin = ex[1]
            nrange = range(iex-1,-1,-1) if left else range(iex+1,len(extrema))
            for n in nrange:
                if extrema[n][0] == 'min':
                    if extrema[n][1] < vmin: vmin = extrema[n][1]
                elif extrema[n][0] == 'max':
                    if extrema[n][1] >= ex[1]:
                        prom_one = ex[1] - vmin
                        break
            return prom_one
        # the total prominence is the minimum of the two
        prom = min(find_one_sided_prominence(iex,ex,left=True),
                   find_one_sided_prominence(iex,ex,left=False))
        if np.isinf(prom):
            prom = arr.ptp()
        ex.append( prom )
        # find the isolation
        iso = max( ex[2]+1, len(arr)-ex[2] )
        for i in range(1,max(ex[2],len(arr)-ex[2])):
            if ex[2]-i>= 0 and arr[ex[2]-i] > ex[1]:
                iso = i
                break
            if ex[2]+i<len(arr) and arr[ex[2]+i] > ex[1]:
                iso = i
                break
        ex.append( iso )

    # filter maxima with given prominence
    maxima = [ ex for ex in extrema
                if (ex[0]=='max' and (prominence is None or ex[3]>prominence)) ]
    if sortby == 'index':
        if descending:
            maxima = sorted(maxima, key=lambda ex: ex[2], reverse=descending)
        else:
            pass    # nothing to do...
    elif sortby == 'value':
        maxima = sorted(maxima, key=lambda ex: ex[1], reverse=descending)
    elif sortby == 'prominence':
        maxima = sorted(maxima, key=lambda ex: ex[3], reverse=descending)
    elif sortby == 'isolation':
        maxima = sorted(maxima, key=lambda ex: ex[4], reverse=descending)
    else:
        raise ValueError('Cannot sort by "%s" -  unknown property.' % sortby)
    return np.array( [(idx,val,prom,iso) for _,val,idx,prom,iso in maxima],
                      dtype=[('index',int), ('value',arr.dtype),
                             ('prominence',arr.dtype), ('isolation',int)] )

